"""
"""
from __future__ import annotations

import threading
from typing import Any
from typing import Callable


reload_server_ports: set[int] = set()


def start_reload_server(
    prerun: Callable[[], Any],
    postrun: Callable[[], bool],
    stop_event: threading.Event,
    port: int = 7887,
):
    from .server import ReloadServer

    reload_server = ReloadServer(
        prerun=prerun,
        postrun=postrun,
        stop_event=stop_event,
    )

    reload_server_ports.add(port)
    try:
        reload_server.run(port=port)
    finally:
        reload_server_ports.remove(port)



"""
"""

from packaging import version

from ...config import Config


try:

    import torch

except ImportError:

    _patch = lambda *args, **kwargs: None
    _unpatch = lambda *args, **kwargs: None
    _pack = lambda *args, **kwargs: 0
    _init = lambda *args, **kwargs: None
    _size = lambda *args, **kwargs: 0
    _move = lambda *args, **kwargs: None
    _is_in_bad_fork = lambda *args, **kwargs: False

else:

    # Make OMP_NUM_THREADS impact both inter-op and intra-op
    if version.parse(torch.__version__) >= version.parse('2.4'):
        num_threads = torch.get_num_threads()
        try:
            torch.set_num_interop_threads(num_threads)
        except RuntimeError: # Already set
            pass

    if Config.zero_gpu_v2:
        from . import patching as _patching
    else: # pragma: no cover
        from . import patching_legacy as _patching

    _patch = _patching.patch
    _unpatch = _patching.unpatch
    _pack = _patching.pack
    _init = _patching.init
    _size = _patching.size
    _move = _patching.move
    _is_in_bad_fork = _patching.is_in_bad_fork


patch = _patch
unpatch = _unpatch
pack = _pack
init = _init
size = _size
move = _move
is_in_bad_fork = _is_in_bad_fork



"""
"""
from __future__ import annotations

import sys
from types import FunctionType
from typing import TYPE_CHECKING
from typing import Callable


if sys.version_info.minor < 8: # pragma: no cover
    raise RuntimeError("Importing PySpaces requires Python 3.8+")


# Prevent gradio from importing spaces
if (gr := sys.modules.get('gradio')) is not None: # pragma: no cover
    try:
        gr.Blocks
    except AttributeError:
        raise ImportError


from .zero.decorator import GPU
from .gradio import gradio_auto_wrap
from .gradio import disable_gradio_auto_wrap
from .gradio import enable_gradio_auto_wrap


class LazyImported:
    def __init__(self, import_fn: Callable[[], FunctionType]):
        self.import_fn = import_fn
    def __call__(self, *args, **kwargs):
        return self.import_fn()(*args, **kwargs)
    @property
    def __wrapped__(self):
        return self.import_fn()
    @property
    def __doc__(self): # pyright: ignore[reportIncompatibleVariableOverride]
        return self.import_fn().__doc__
    @property
    def __code__(self):
        return self.import_fn.__code__
    @property
    def __class__(self): # pyright: ignore[reportIncompatibleMethodOverride]
        return FunctionType
    @property
    def __name__(self):
        return self.import_fn.__name__.removeprefix('_')


def _aoti_capture():
    from .zero.torch.aoti import aoti_capture
    return aoti_capture

def _aoti_compile():
    from .zero.torch.aoti import aoti_compile
    return aoti_compile

def _aoti_apply():
    from .zero.torch.aoti import aoti_apply
    return aoti_apply

def _aoti_blocks_load():
    from .zero.torch.aoti import aoti_blocks_load
    return aoti_blocks_load


aoti_capture = LazyImported(_aoti_capture)
aoti_compile = LazyImported(_aoti_compile)
aoti_apply = LazyImported(_aoti_apply)
aoti_blocks_load = LazyImported(_aoti_blocks_load)


if TYPE_CHECKING:
    from .zero.torch.aoti import aoti_capture
    from .zero.torch.aoti import aoti_compile
    from .zero.torch.aoti import aoti_apply
    from .zero.torch.aoti import aoti_blocks_load


__all__ = [
    'GPU',
    'gradio_auto_wrap',
    'disable_gradio_auto_wrap',
    'enable_gradio_auto_wrap',
    'aoti_capture',
    'aoti_compile',
    'aoti_apply',
    'aoti_blocks_load',
]



import importlib.util
import logging
import os
import sys
from _frozen_importlib_external import SourceFileLoader
from types import CodeType, FunctionType, ModuleType

from ovld import OvldMC, ovld

from .codetools import CodeFile, FunctionDefinition
from .utils import EventSource, glob_filter

log = logging.getLogger(__name__)


class Registry(metaclass=OvldMC):
    def __init__(self):
        self.filename_to_module = {}
        # Cache of (module_name, file_contents, mtime)
        # A snapshot of the file contents may be saved before it might be modified
        self.precache = {}
        # Cache of CodeFile (lazy)
        self.cache = {}
        self.precache_activity = EventSource(save_history=True)
        self.activity = EventSource()
        self._log = None

    def set_logger(self, log):
        self._log = log

    def log(self, *args, **kwargs):
        if self._log is not None:
            self._log(*args, **kwargs)

    def prepare(self, module_name=None, filename=None):
        if filename is None:
            assert module_name is not None
            filename = sys.modules[module_name].__file__

        if filename not in self.precache and filename not in self.cache:
            if module_name is None:
                if filename in self.filename_to_module:
                    module_name = self.filename_to_module[filename]
                else:
                    for module_name, module in sys.modules.items():
                        fname = getattr(module, "__file__", None)
                        if fname:
                            self.filename_to_module[fname] = module_name
                            if fname == filename:
                                break
                    else:  # pragma: no cover
                        raise Exception(
                            f"Cannot find module that corresponds to {filename}"
                        )

            if os.path.exists(filename):
                with open(filename, "r", encoding="utf8") as f:
                    self.precache[filename] = (
                        module_name,
                        f.read(),
                        os.path.getmtime(filename),
                    )
                self.precache_activity.emit(module_name, filename)

        return module_name, filename

    def get(self, filename):
        if filename in self.cache:
            return self.cache[filename]

        if filename in self.precache:
            module_name, cached_source, mtime = self.precache[filename]
            if module_name not in sys.modules:
                return None
            cf = CodeFile(
                filename, source=cached_source, module_name=module_name
            )
            cf.associate(sys.modules[module_name])
            cf.activity.register(self.log)
            # Basic forwarding of the CodeFile's events
            cf.activity.register(self.activity.emit)
            self.cache[filename] = cf
            return cf

        return None

    def get_at(self, filename, lineno):
        cf = self.get(filename)
        if cf is None:
            return None, None
        for entry in cf.root.walk():
            if (
                isinstance(entry, FunctionDefinition)
                and entry.node is not None
                and (
                    (
                        entry.stashed.lineno == lineno
                        and entry.stashed.filename == filename
                    )
                    or (
                        entry.node.extent.lineno == lineno
                        and entry.node.extent.filename == filename
                    )
                )
            ):
                return cf, entry
        else:
            return cf, None

    def auto_register(self, filter=glob_filter("./*.py")):
        def prep(module_name, filename):
            if (
                filename is not None
                and module_name is not None
                and filter(filename)
            ):
                try:
                    self.prepare(module_name, filename)
                except (UnicodeDecodeError, OSError):  # pragma: no cover
                    pass

        for name, module in list(sys.modules.items()):
            filename = getattr(module, "__file__", None)
            module_name = getattr(module, "__name__", None)
            prep(module_name, filename)

        return add_sniffer(prep)

    @ovld
    def find(self, module: ModuleType):
        self.prepare(module.__name__, module.__file__)
        cf = self.get(module.__file__)
        return cf, cf.root

    @ovld
    def find(self, fn: FunctionType):
        co = fn.__code__
        self.prepare(fn.__module__, co.co_filename)
        return self.get_at(co.co_filename, co.co_firstlineno)

    @ovld
    def find(self, co: CodeType):
        self.prepare(filename=co.co_filename)
        return self.get_at(co.co_filename, co.co_firstlineno)

    @ovld
    def find(self, cls: type):
        _, filename = self.prepare(module_name=cls.__module__)
        cf = self.get(filename)
        key = f"{cls.__module__}.{cls.__qualname__}"
        for entry in cf.root.walk():
            if entry.dotpath() == key:
                return cf, entry
        else:
            return cf, None


registry = Registry()


class ImportSniffer:
    """A spec finder that simply sniffs for attempted imports.

    Basically we install this at the front of sys.meta_path so that
    importlib.util.find_spec calls it, then we call find_spec
    ourselves to locate the file that's going to be read so that we
    know we have to cache its contents and watch for changes.
    """

    def __init__(self):
        self.working = False

    def find_spec(self, fullname, path, target=None):
        if not _sniffer_callbacks:
            return None

        if not self.working:
            self.working = True
            # We call find_spec ourselves to find out where the file is.
            # This will not cause an infinite loop because self.working
            # is True and we will not enter the conditional. I'm not
            # sure if it's dangerous to call find_spec within find_spec,
            # but it seems to work, so whatever.
            mspec = importlib.util.find_spec(fullname, path)
            if (
                mspec is not None
                and isinstance(mspec.loader, SourceFileLoader)
                and mspec.name is not None
                and mspec.origin is not None
            ):
                for report in _sniffer_callbacks:
                    try:
                        report(mspec.name, mspec.origin)
                    except Exception as exc:
                        log.error(
                            f"jurigged: Error processing spec {mspec.name}",
                            exc_info=exc,
                        )
            self.working = False
        return None


_main_sniffer = ImportSniffer()
sys.meta_path.insert(0, _main_sniffer)
_sniffer_callbacks = []


def add_sniffer(report):
    _sniffer_callbacks.append(report)
    report.uninstall = lambda: _sniffer_callbacks.remove(report)
    return report



"""
"""

import httpx

from .._vendor.sseclient import SSEClient

from .types import ApiGetStatusRequest
from .types import ApiGetStatusResponse
from .types import ApiFetchContentsRequest
from .types import ApiFetchContentsResponse
from .types import ApiCreateReloadRequest
from .types import ApiCreateReloadResponse
from .types import ApiGetReloadEventSourceData
from .types import ApiGetReloadRequest


class ReloadClient:
    def __init__(self, http_client: httpx.Client):
        self.client = http_client

    def get_status(self, revision: str): # pragma: no cover
        req = ApiGetStatusRequest(revision=revision)
        res = self.client.post('/get-status', json=req.model_dump())
        assert res.status_code == 200, res.status_code
        return ApiGetStatusResponse.model_validate(res.json())

    def fetch_contents(self, filepath: str):
        req = ApiFetchContentsRequest(filepath=filepath)
        res = self.client.post('/fetch-contents', json=req.model_dump())
        assert res.status_code == 200, res.status_code
        return ApiFetchContentsResponse.model_validate(res.json())

    def create_reload(self, filepath: str, contents: str):
        req = ApiCreateReloadRequest(filepath=filepath, contents=contents)
        res = self.client.post('/create-reload', json=req.model_dump())
        assert res.status_code == 200, res.status_code
        return ApiCreateReloadResponse.model_validate(res.json())

    def get_reload(self, reload_id: str):
        req = ApiGetReloadRequest(reloadId=reload_id)
        with self.client.stream('POST', '/get-reload', json=req.model_dump()) as res:
            assert res.status_code == 200, res.status_code
            for event in SSEClient(res.iter_bytes()).events():
                if event.event == 'message':
                    yield ApiGetReloadEventSourceData.model_validate_json(event.data)



"""
"""
from __future__ import annotations

import os
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Literal
from typing import cast
from typing import get_args

from typing_extensions import TypeAlias
from typing_extensions import assert_type

from .zero.api import GPUSize


GPUSizeConfig = Literal['auto', 'medium', 'large']


ZEROGPU_HOME = Path.home() / '.zerogpu'


def boolean(value: str | None) -> bool:
    return value is not None and value.lower() in ("1", "t", "true")


class Settings:
    def __init__(self):
        self.zero_gpu = boolean(
            os.getenv('SPACES_ZERO_GPU'))
        self.zero_device_api_url = (
            os.getenv('SPACES_ZERO_DEVICE_API_URL'))
        self.gradio_auto_wrap = boolean(
            os.getenv('SPACES_GRADIO_AUTO_WRAP'))
        self.zero_patch_torch_device = boolean(
            os.getenv('ZERO_GPU_PATCH_TORCH_DEVICE'))
        self.zero_gpu_v2 = boolean(
            os.getenv('ZEROGPU_V2', "true"))
        self.zerogpu_size: GPUSize | Literal['auto'] = cast(GPUSizeConfig,
            os.getenv('ZEROGPU_SIZE', 'large'))
        self.zerogpu_medium_size_threshold = int(
            os.getenv('ZEROGPU_MEDIUM_SIZE_THRESHOLD', 30 * 2**30))
        self.zerogpu_offload_dir = (
            os.getenv('ZEROGPU_OFFLOAD_DIR', str(ZEROGPU_HOME / 'tensors')))
        self.zerogpu_proc_self_cgroup_path = (
            os.getenv('ZEROGPU_PROC_SELF_CGROUP_PATH', '/proc/self/cgroup'))
        self.zerogpu_cuda_device_name = (
            os.getenv('ZEROGPU_CUDA_DEVICE_NAME', "NVIDIA H200 MIG 3g.71gb"))
        self.zerogpu_cuda_total_memory = int(
            os.getenv('ZEROGPU_CUDA_TOTAL_MEMORY', 74625056768))
        self.zerogpu_cuda_reserved_memory = int(
            os.getenv('ZEROGPU_CUDA_RESERVED_MEMORY', 0))
        self.zerogpu_cuda_capability_major = int(
            os.getenv('ZEROGPU_CUDA_CAPABILITY_MAJOR', 9))
        self.zerogpu_cuda_capability_minor = int(
            os.getenv('ZEROGPU_CUDA_CAPABILITY_MINOR', 0))
        self.zerogpu_cuda_multi_processor_count = int(
            os.getenv('ZEROGPU_CUDA_MULTI_PROCESSOR_COUNT', 60))


Config = Settings()


if TYPE_CHECKING:
    assert_type(Config.zerogpu_size, GPUSizeConfig)


if Config.zero_gpu:
    assert Config.zero_device_api_url is not None, (
        'SPACES_ZERO_DEVICE_API_URL env must be set '
        'on ZeroGPU Spaces (identified by SPACES_ZERO_GPU=true)'
    )
    assert Config.zerogpu_size in (values := get_args(GPUSizeConfig)), (
        f"ZEROGPU_SIZE should be one of {', '.join(values)}"
    )



"""
"""
# pyright: reportPrivateImportUsage=false

from __future__ import annotations

import gc
import multiprocessing
import os
import shutil
from collections import defaultdict
from concurrent.futures import ProcessPoolExecutor
from concurrent.futures import ThreadPoolExecutor
from contextlib import nullcontext
from contextvars import copy_context
from pathlib import Path
from typing import Any
from typing import Callable

import torch
from torch.overrides import TorchFunctionMode
from torch.overrides import resolve_name
from torch.utils._python_dispatch import is_traceable_wrapper_subclass
from torch.utils._python_dispatch import transform_subclass
from torch.utils._python_dispatch import TorchDispatchMode
from torch.utils._pytree import tree_map_only
from torch.utils.weak import WeakTensorKeyDictionary

from ...config import Config
from ..tqdm import tqdm
from ..utils import malloc_trim
from . import cudart
from .packing import ZeroGPUTensorPack
from .packing import pack_tensors
from .packing import pack_to_cuda
from .static import *
from .utils import empty_like_raw_alloc
from .types import AliasId


PINNED_MEMORY_RATIO_LIMIT = 0.1

OPS_INPUTS_CHECK_NO_RETURN = (
    torch.Tensor.equal,
)

OPS_INPUT_CHECK_SELF_RETURN = (
    torch.Tensor.set_, # probably never dispatched
    torch.ops.aten.set_.source_Tensor, # pyright: ignore [reportAttributeAccessIssue]
)

OFFLOADED_ERROR_MESSAGE = "Cannot apply function {} on disk-offloaded Tensor {}"

_tensor_make_subclass = torch.Tensor._make_subclass
_asarray           = torch.asarray
_device            = torch.device
_cuda_init         = torch._C._cuda_init
_cuda_exchange_device = torch.cuda._exchange_device
_cuda_available      = torch.cuda.is_available
_cuda_device_count   = torch.cuda.device_count
_cuda_current_device = torch.cuda.current_device
_cuda_synchronize    = torch.cuda.synchronize
_cuda_get_device_capability   = torch.cuda.get_device_capability
_cuda_get_device_properties   = torch.cuda.get_device_properties
_cuda_get_device_name         = torch.cuda.get_device_name
_cuda_memory_stats_as_nested_dict = torch.cuda.memory.memory_stats_as_nested_dict
_cuda_cudart = torch.cuda.cudart

# PyTorch 2.3
_cuda_maybe_exchange_device = getattr(torch.cuda, '_maybe_exchange_device', None)


cuda_aliases: dict[torch.Tensor, torch.Tensor | None] = WeakTensorKeyDictionary() # pyright: ignore [reportAssignmentType]

tensor_packs: list[ZeroGPUTensorPack] = []

class ZeroGPUTensor(torch.Tensor):
    pass

def empty_fake(tensor: torch.Tensor):
    fake = empty_like_raw_alloc(tensor, requires_grad=tensor.requires_grad)
    if fake.__class__ != tensor.__class__:
        fake = _tensor_make_subclass(tensor.__class__, fake, require_grad=tensor.requires_grad) # pyright: ignore [reportArgumentType]
    return fake

# Torch 2.5: https://github.com/pytorch/pytorch/issues/144152
def no_int_device(*args, **kwargs):
    if len(args) and isinstance(index := args[0], int):
        args = (f'cuda:{index}', *args[1:])
    if isinstance(index := kwargs.get('device'), int):
        kwargs['device'] = f'cuda:{index}'
    return args, kwargs


class ZeroGPUFunctionMode(TorchFunctionMode):

    def __torch_function__(self, func, types, args=(), kwargs: dict[str, Any] | None = None):

        kwargs = {} if kwargs is None else kwargs

        if func == torch._C._nn._parse_to: # pyright: ignore [reportAttributeAccessIssue]
            args, kwargs = no_int_device(*args, **kwargs)
            return func(*args, **kwargs)

        # Redispatch: tensor.cuda() -> tensor.to(device='cuda')
        if func == torch.Tensor.cuda or func == torch.Tensor.cpu:
            memory_format = kwargs.get('memory_format')
            return self.__torch_function__(torch.Tensor.to, types, (args[0],), {
                'device': 'cuda' if func == torch.Tensor.cuda else 'cpu',
                **({'memory_format': memory_format} if memory_format is not None else {}),
            })

        # Redispatch: tensor.to('cuda') -> tensor.to(device='cuda')
        if func == torch.Tensor.to and len(args) > 1:
            parse_to_args, parse_to_kwargs = no_int_device(*args[1:], **kwargs)
            # We are using nn._parse_to utility to parse generic Tensor.to but nn does not accept copy kwarg
            copy_kwarg = {'copy': parse_to_kwargs.pop('copy')} if 'copy' in parse_to_kwargs else {}
            device, dtype, _, memory_format = torch._C._nn._parse_to(*parse_to_args, **parse_to_kwargs) # pyright: ignore [reportAttributeAccessIssue]
            return self.__torch_function__(torch.Tensor.to, types, (args[0],), {
                'device': device,
                'dtype': dtype,
                'memory_format': memory_format,
            } | copy_kwarg)

        if func == torch.Tensor.data.__set__: # pyright: ignore [reportAttributeAccessIssue]
            self, target = args
            if target in cuda_aliases:
                if (target_original := cuda_aliases[target]) is None:
                    raise Exception(OFFLOADED_ERROR_MESSAGE.format(resolve_name(func), target))
                original = empty_fake(self)
                original.data = target_original
                cuda_aliases[self] = original
            elif self in cuda_aliases:
                del cuda_aliases[self]
            self.data = target
            return

        if func == torch.Tensor.device.__get__:
            tensor, = args
            if tensor in cuda_aliases:
                return torch.device('cuda', index=0)

        elif func == torch.Tensor.__repr__:
            tensor, = args
            if tensor in cuda_aliases:
                if (original := cuda_aliases[tensor]) is None:
                    original = tensor.to('meta')
                original_class = original.__class__
                original.__class__ = ZeroGPUTensor
                try:
                    return func(original, **kwargs)
                finally:
                    original.__class__ = original_class

        elif func == torch.Tensor.untyped_storage:
            tensor, = args
            if tensor in cuda_aliases:
                if (original := cuda_aliases[tensor]) is None:
                    raise Exception(OFFLOADED_ERROR_MESSAGE.format(resolve_name(func), tensor))
                res = func(original, **kwargs)
                res._zerogpu = True
                return res

        cuda: bool | None = None

        # Handle device kwarg
        if (device := kwargs.get('device')) is not None:
            device = torch.device(device)
            if device.type == 'cuda':
                kwargs['device'] = torch.device('cpu')
                cuda = True
            else:
                cuda = False

        # Swap fake inputs with original data
        swapped = {}
        inputs_are_cuda = set()
        def swap(tensor: torch.Tensor):
            nonlocal inputs_are_cuda
            if tensor not in cuda_aliases:
                inputs_are_cuda |= {False}
                return tensor
            if (original := cuda_aliases[tensor]) is None:
                raise Exception(OFFLOADED_ERROR_MESSAGE.format(resolve_name(func), tensor))
            swapped[original] = tensor
            inputs_are_cuda |= {True}
            return original
        args_ = tree_map_only(torch.Tensor, swap, args)
        kwargs_ = tree_map_only(torch.Tensor, swap, kwargs)
        if inputs_are_cuda == {True}:
            if cuda is not False:
                cuda = True

        # Wrapper tensors special case (torchao quickix)
        if len(args) == 1 and is_traceable_wrapper_subclass(wrapper_tensor := args[0]):
            if func in {
                torch.Tensor.detach,
                torch.ops.aten.alias.default, # pyright: ignore [reportAttributeAccessIssue]
                torch.ops.aten.clone.default, # pyright: ignore [reportAttributeAccessIssue]
            }:
                with self:
                    return transform_subclass(wrapper_tensor, lambda _, t: func(t))

        res = func(*args_, **kwargs_)

        # Re-generate swapped fakes in case of mutation
        for original, fake in swapped.items():
            fake.data = empty_fake(original)

        # Special case for Tensor indexing where only 'self' matters
        if func in {
            torch.ops.aten.index.Tensor, # pyright: ignore [reportAttributeAccessIssue]
            torch.Tensor.__getitem__, # PyTorch 2.4+
        }:
            self = args[0]
            cuda = self in cuda_aliases
            inputs_are_cuda = {cuda}

        # Emulate device check
        if isinstance(res, torch.Tensor) or func in OPS_INPUTS_CHECK_NO_RETURN:
            self = None
            if len(args_) >= 1 and isinstance(args_[0], torch.Tensor):
                self = args_[0]
            # Only raise if func does not return its first input (Tensor.copy_)
            if res is not self or func in OPS_INPUT_CHECK_SELF_RETURN:
                if inputs_are_cuda == {True, False}:
                    raise RuntimeError(
                        "Expected all tensors to be on the same device, "
                        "but found at least two devices, cuda:0 (ZeroGPU) and cpu!"
                    )

        # Register output
        def register(tensor: torch.Tensor):
            if tensor in swapped and cuda is not False:
                return swapped[tensor]
            if cuda is not True:
                return tensor
            fake = empty_fake(tensor)
            cuda_aliases[fake] = tensor
            return fake

        return tree_map_only(torch.Tensor, register, res)

# When enabling DispatchMode, some aten ops are dispatched to FunctionMode
# We are using it for aten.alias.default and aten.set_.source_Tensor
class DefaultDispatchMode(TorchDispatchMode):
    def __torch_dispatch__(self, func, types, args=(), kwargs: dict[str, Any] | None = None):
        return func(*args, **(kwargs or {}))


function_mode = ZeroGPUFunctionMode()
dispatch_mode = DefaultDispatchMode()


def _untyped_storage_new_register(*args, **kwargs):
    cuda = False
    if (device := kwargs.get('device')) is not None:
        device = torch.device(device)
        if device.type == 'cuda':
            cuda = True
            del kwargs['device']
    storage = torch._C.StorageBase.__new__(*args, **kwargs)
    if cuda:
        storage._zerogpu = True
    return storage

@property
def _untyped_storage_device(self):
    if hasattr(self, '_zerogpu'):
        return torch.device('cuda', index=0)
    return torch._C.StorageBase.device.__get__(self) # pyright: ignore [reportAttributeAccessIssue]

# Force dispatch
def _tensor_make_subclass_function_mode(*args, **kwargs):
    with torch._C.DisableTorchFunction():
        return function_mode.__torch_function__(_tensor_make_subclass, (), args=args, kwargs=kwargs)
def _asarray_function_mode(*args, **kwargs):
    with torch._C.DisableTorchFunction():
        return function_mode.__torch_function__(_asarray, (), args=args, kwargs=kwargs)

class _DeviceStringOnlyMeta(type):
    def __instancecheck__(cls, instance):
        return isinstance(instance, _device)

class _DeviceStringOnly(metaclass=_DeviceStringOnlyMeta):
    def __new__(cls, *args, **kwargs):
        args, kwargs = no_int_device(*args, **kwargs)
        return _device(*args, **kwargs)

def _cuda_init_raise():
    raise RuntimeError(
        "CUDA must not be initialized in the main process "
        "on Spaces with Stateless GPU environment.\n"
        "You can look at this Stacktrace to find out "
        "which part of your code triggered a CUDA init"
    )

def _cuda_dummy_exchange_device(device):
    assert device in {-1, 0}
    return device

def patch():
    function_mode.__enter__()
    dispatch_mode.__enter__()
    # TODO: only patch bellow methods on current Thread to be consistent with TorchModes
    # (or hijack threading.Thread.__init__ to force Modes on all threads)
    torch.Tensor._make_subclass = _tensor_make_subclass_function_mode # pyright: ignore [reportAttributeAccessIssue]
    torch.UntypedStorage.__new__ = _untyped_storage_new_register
    torch.UntypedStorage.device  = _untyped_storage_device # pyright: ignore [reportAttributeAccessIssue]
    torch.asarray           = _asarray_function_mode
    torch.device            = _DeviceStringOnly
    torch._C._cuda_init     = _cuda_init_raise
    torch.cuda._exchange_device = _cuda_dummy_exchange_device
    torch.cuda.is_available   = lambda: True
    torch.cuda.device_count   = lambda: 1
    torch.cuda.current_device = lambda: 0
    torch.cuda.synchronize    = lambda *args: None
    torch.cuda.get_device_capability = lambda *args, **kwargs: CUDA_DEVICE_CAPABILITY
    torch.cuda.get_device_properties = lambda *args, **kwargs: CUDA_DEVICE_PROPERTIES
    torch.cuda.get_device_name       = lambda *args, **kwargs: CUDA_DEVICE_NAME
    torch.cuda.memory.memory_stats_as_nested_dict = lambda *args, **kwargs: CUDA_MEMORY_STATS_AS_NESTED_DICT
    torch.cuda.cudart = lambda: cudart
    # PyTorch 2.3
    if _cuda_maybe_exchange_device is not None: # pragma: no cover
        setattr(torch.cuda, '_maybe_exchange_device', _cuda_dummy_exchange_device)
    bitsandbytes().patch()

def unpatch():
    try:
        dispatch_mode.__exit__(None, None, None)
        function_mode.__exit__(None, None, None)
    except RuntimeError:
        pass # patch() and unpatch() called from != threads
    torch.Tensor._make_subclass = _tensor_make_subclass
    torch.UntypedStorage.__new__ = torch._C.StorageBase.__new__
    torch.UntypedStorage.device  = torch._C.StorageBase.device # pyright: ignore [reportAttributeAccessIssue]
    torch.asarray           = _asarray
    torch.device            = _device
    torch._C._cuda_init     = _cuda_init
    torch.cuda._exchange_device = _cuda_exchange_device
    torch.cuda.is_available   = _cuda_available
    torch.cuda.device_count   = _cuda_device_count
    torch.cuda.current_device = _cuda_current_device
    torch.cuda.synchronize    = _cuda_synchronize
    torch.cuda.get_device_capability = _cuda_get_device_capability
    torch.cuda.get_device_properties = _cuda_get_device_properties
    torch.cuda.get_device_name       = _cuda_get_device_name
    torch.cuda.memory.memory_stats_as_nested_dict = _cuda_memory_stats_as_nested_dict
    torch.cuda.cudart = _cuda_cudart
    # PyTorch 2.3
    if _cuda_maybe_exchange_device is not None: # pragma: no cover
        setattr(torch.cuda, '_maybe_exchange_device', _cuda_exchange_device)
    bitsandbytes().unpatch()


def _total_unpacked_size():
    tensors = [tensor for tensor in cuda_aliases.values() if tensor is not None]
    deduped = {AliasId.from_tensor(tensor): tensor for tensor in tensors}
    return sum([tensor.numel() * tensor.element_size() for tensor in deduped.values()])


def _pack(offload_dir: str):
    # Pack to disk
    originals: set[torch.Tensor] = set()
    originals_dedup: dict[AliasId, torch.Tensor] = {}
    fakes: dict[torch.Tensor, list[torch.Tensor]] = defaultdict(list)
    for fake, original in cuda_aliases.items():
        # TODO filter-out sparse Tensors
        if original is not None:
            original_id = AliasId.from_tensor(original)
            if original_id not in originals_dedup:
                originals_dedup[original_id] = original
                originals |= {original}
            fakes[originals_dedup[original_id]] += [fake]
    total_size = _total_unpacked_size()
    progress = tqdm(
        total=total_size,
        unit='B',
        unit_scale=True,
        desc="ZeroGPU tensors packing",
    ) if tqdm is not None else nullcontext()
    with progress as progress:
        update = progress.update if progress is not None else lambda _: None
        pack = pack_tensors(originals, fakes, offload_dir, callback=update)
    tensor_packs.append(pack)
    # Free memory
    for fake_list in fakes.values():
        for fake in fake_list:
            cuda_aliases[fake] = None
    return total_size

def pack():
    shutil.rmtree(Config.zerogpu_offload_dir, ignore_errors=True)
    Path(Config.zerogpu_offload_dir).mkdir(parents=True)
    total_size = _pack(Config.zerogpu_offload_dir)
    gc.collect()
    malloc_trim()
    return total_size

def init(nvidia_uuid: str):
    os.environ['CUDA_VISIBLE_DEVICES'] = nvidia_uuid
    torch.Tensor([0]).cuda()

def size():
    return _total_unpacked_size() + sum([pack.total_size for pack in tensor_packs])

def _move(callback: Callable[[int]] | None = None):
    callback = callback if callback is not None else lambda _: None
    # CPU -> CUDA
    pinned_limit = _total_unpacked_size() * PINNED_MEMORY_RATIO_LIMIT
    moved: dict[AliasId, torch.Tensor] = {}
    for fake, original in cuda_aliases.items():
        if original is not None:
            original = torch.Tensor(original) # unwrap subclass
            original_id = AliasId.from_tensor(original)
            if original_id not in moved:
                if original.numel() * original.element_size() < pinned_limit:
                    original_cuda = original.pin_memory().cuda(non_blocking=True)
                else:
                    original_cuda = original.cuda()
                moved[original_id] = original_cuda
                callback(fake.numel() * fake.element_size())
    torch.cuda.synchronize()
    for fake, original in cuda_aliases.items():
        if original is not None:
            fake.data = moved[AliasId.from_tensor(original)]
    # Disk -> CUDA
    for tensor_pack in tensor_packs:
        pack_to_cuda(tensor_pack, callback=callback)
    bitsandbytes().move()

def move(callback: Callable[[int]] | None = None):
    callback = callback if callback is not None else lambda _: None
    with ThreadPoolExecutor(1) as e:
        e.submit(copy_context().run, _move, callback=callback).result()
    torch.cuda.synchronize()

def is_in_bad_fork():
    with ProcessPoolExecutor(mp_context=multiprocessing.get_context('fork')) as e:
        f = e.submit(torch.cuda._is_in_bad_fork)
        return f.result()

def bitsandbytes():
    # Lazy import
    from . import bitsandbytes
    return bitsandbytes



"""
"""
from __future__ import annotations


from dataclasses import dataclass
from datetime import timedelta
from typing import Any
from typing import Dict
from typing import Tuple
from typing import TypedDict
from typing_extensions import Callable
from typing_extensions import Generic
from typing_extensions import ParamSpec
from typing_extensions import TypeAlias
from typing_extensions import TypeVar


Params = Tuple[Tuple[object, ...], Dict[str, Any]]
Res = TypeVar('Res')
Param = ParamSpec('Param')

class EmptyKwargs(TypedDict):
    pass

@dataclass
class OkResult(Generic[Res]):
    value: Res
@dataclass
class ExceptionResult:
    traceback: str
    error_cls: str
@dataclass
class AbortedResult:
    pass
@dataclass
class EndResult:
    pass
@dataclass
class GradioQueueEvent:
    method_name: str
    args: tuple[Any, ...]
    kwargs: dict[str, Any]

RegularResQueueResult:   TypeAlias = "OkResult[Res] | ExceptionResult | GradioQueueEvent"
GeneratorResQueueResult: TypeAlias = "OkResult[Res] | ExceptionResult | EndResult | GradioQueueEvent"
YieldQueueResult:        TypeAlias = "OkResult[Res] | ExceptionResult | EndResult | AbortedResult"

Duration:        TypeAlias = "int | timedelta"
DynamicDuration: TypeAlias = "Duration | Callable[Param, Duration] | None"



"""
"""

import mmap
from functools import wraps

import torch


@wraps(torch.empty_like)
def empty_like_raw_alloc(tensor: torch.Tensor, **kwargs) -> torch.Tensor:
    empty = torch.empty_like(tensor, **{**kwargs, 'requires_grad': False})
    if (nbytes := empty.untyped_storage().nbytes()) > 0:
        buffer = mmap.mmap(-1, nbytes, prot=mmap.PROT_READ | mmap.PROT_WRITE)
        buffer = torch.frombuffer(buffer, dtype=torch.uint8)
        empty.set_(buffer.untyped_storage(), 0, empty.shape, empty.stride())
    empty.requires_grad_(kwargs.get('requires_grad', False))
    return empty



import fnmatch
import os
import types


class EventSource(list):
    def __init__(self, *, save_history=False):
        if save_history:
            self._history = []
        else:
            self._history = None

    def register(self, listener, apply_history=True):
        if self._history and apply_history:
            for args, kwargs in self._history:
                listener(*args, **kwargs)
        self.append(listener)
        return listener

    def emit(self, *args, **kwargs):
        for listener in self:
            listener(*args, **kwargs)
        if self._history is not None:
            self._history.append((args, kwargs))


def glob_filter(pattern):
    if pattern.startswith("~"):
        pattern = os.path.expanduser(pattern)
    elif not pattern.startswith("/"):
        pattern = os.path.abspath(pattern)

    if os.path.isdir(pattern):
        pattern = os.path.join(pattern, "*")

    def matcher(filename):
        return fnmatch.fnmatch(filename, pattern)

    return matcher


def or_filter(filters):
    if len(filters) == 1:
        return filters[0]

    def matcher(filename):
        return any(f(filename) for f in filters)

    return matcher


def shift_lineno(co, delta):
    if isinstance(co, types.CodeType):
        return co.replace(
            co_firstlineno=co.co_firstlineno + delta,
            co_consts=tuple(shift_lineno(ct, delta) for ct in co.co_consts),
        )
    else:
        return co



from .registry import CodeRegistry, ConformException  # noqa: F401
from .version import version  # noqa: F401

code_registry = CodeRegistry()
code_registry.setup()

collect_all = code_registry.collect_all
conform = code_registry.conform
find_code = code_registry.find_code
get_functions = code_registry.get_functions
update_cache_entry = code_registry.update_cache_entry



"""
"""
# pyright: reportPrivateImportUsage=false

from __future__ import annotations

import importlib
from contextlib import contextmanager
from contextlib import nullcontext
from importlib import metadata
from types import ModuleType
from typing import TYPE_CHECKING
from typing import Tuple

import torch
from packaging import version

if TYPE_CHECKING:
    import torch as Torch


@contextmanager
def cuda_unavailable(torch: ModuleType): # pragma: no cover
    _is_available = torch.cuda.is_available
    torch.cuda.is_available = lambda: False
    yield
    torch.cuda.is_available = _is_available


def maybe_import_bitsandbytes():
    try:
        import torch
    except ImportError: # pragma: no cover
        return None
    try:
        bnb_version = version.parse(metadata.version('bitsandbytes'))
    except ImportError: # pragma: no cover
        return None
    if bnb_version < version.parse('0.40.0'): # pragma: no cover
        raise RuntimeError(f"ZeroGPU requires bitsandbytes >= 0.40.0 (installed: {bnb_version})")
    if bnb_version < version.parse('0.43.1'): # pragma: no cover
        context = lambda: cuda_unavailable(torch)
    else:
        context = lambda: nullcontext()
    with (ctx := context()):
        try:
            import bitsandbytes
        except ImportError:
            return None
        if not isinstance(ctx, nullcontext): # pragma: no cover
            print("↑ Those bitsandbytes warnings are expected on ZeroGPU ↑")
    return context


if (import_context := maybe_import_bitsandbytes()):

    from torch.utils.weak import WeakTensorKeyDictionary

    with (import_ctx := import_context()):
        CUDASetup = None
        if not isinstance(import_ctx, nullcontext): # pragma: no cover
            from bitsandbytes.cuda_setup.main import CUDASetup # pyright: ignore [reportMissingImports]
        from bitsandbytes import cextension
        from bitsandbytes import functional
        from bitsandbytes.nn import Int8Params
        from bitsandbytes.nn import Params4bit

    _param_to_8bit   = Int8Params.to     # type: ignore
    _param_cuda_8bit = Int8Params.cuda
    _param_to_4bit   = Params4bit.to     # type: ignore
    _param_cuda_4bit = Params4bit.cuda

    TensorToArgs = Tuple[torch.device, torch.dtype, bool, torch.memory_format]

    to_ops_8bit: dict[Int8Params, TensorToArgs | None] = WeakTensorKeyDictionary() # type: ignore
    to_ops_4bit: dict[Params4bit, TensorToArgs | None] = WeakTensorKeyDictionary() # type: ignore

    def _to_op_register_8bit(self: Int8Params, *args, **kwargs):
        parsed = torch._C._nn._parse_to(*args, **kwargs) # pyright: ignore [reportAttributeAccessIssue]
        device, *_ = parsed
        if not isinstance(device, torch.device): # pragma: no cover
            return _param_to_8bit(self, *args, **kwargs)
        if device.type != 'cuda':
            return _param_to_8bit(self, *args, **kwargs)
        to_ops_8bit[self] = parsed
        return self

    def _to_op_register_4bit(self: Params4bit, *args, **kwargs):
        parsed = torch._C._nn._parse_to(*args, **kwargs) # pyright: ignore [reportAttributeAccessIssue]
        device, *_ = parsed
        if not isinstance(device, torch.device): # pragma: no cover
            return _param_to_4bit(self, *args, **kwargs)
        if device.type != 'cuda':
            return _param_to_4bit(self, *args, **kwargs)
        to_ops_4bit[self] = parsed
        return self

    def _cuda_op_arg_check(device: Torch.device | int | str | None) -> bool:
        if device is None: # pragma: no cover
            return True
        if isinstance(device, int):
            return True
        if isinstance(device, str): # pragma: no cover
            device = torch.device(device)
        return device.type == 'cuda' # pragma: no cover

    def _cuda_op_register_8bit(self: Int8Params, device: Torch.device | int | str | None = None, **kwargs):
        if not _cuda_op_arg_check(device): # pragma: no cover
            # Let PyTorch handle the fail
            return _param_cuda_8bit(self, device, **kwargs)
        to_ops_8bit[self] = None
        return self

    def _cuda_op_register_4bit(self: Params4bit, device: Torch.device | int | str | None = None, **kwargs):
        if not _cuda_op_arg_check(device): # pragma: no cover
            # Let PyTorch handle the fail
            return _param_cuda_4bit(self, device, **kwargs)
        to_ops_4bit[self] = None
        return self

    def _patch():
        Int8Params.to   = _to_op_register_8bit   # type: ignore
        Int8Params.cuda = _cuda_op_register_8bit # type: ignore
        Params4bit.to   = _to_op_register_4bit   # type: ignore
        Params4bit.cuda = _cuda_op_register_4bit # type: ignore

    def _unpatch():
        Int8Params.to   = _param_to_8bit   # type: ignore
        Int8Params.cuda = _param_cuda_8bit
        Params4bit.to   = _param_to_4bit   # type: ignore
        Params4bit.cuda = _param_cuda_4bit

    def _move():
        if CUDASetup is not None: # pragma: no cover
            CUDASetup._instance = None
            importlib.reload(cextension)
            functional.lib = cextension.lib
        for op in to_ops_8bit.items():
            tensor, parsed_args = op
            if parsed_args:
                _, dtype, _, memory_format = parsed_args
            else:
                dtype, memory_format = None, None
            tensor.data = _param_to_8bit(tensor,
                device='cuda',
                dtype=dtype,
                memory_format=memory_format,
            ) # type: ignore
        for op in to_ops_4bit.items():
            tensor, parsed_args = op
            if parsed_args:
                _, dtype, _, memory_format = parsed_args
            else:
                dtype, memory_format = None, None
            tensor.data = _param_to_4bit(tensor,
                device='cuda',
                dtype=dtype,
                memory_format=memory_format,
            ) # type: ignore

else:

    _patch = lambda: None
    _unpatch = lambda: None
    _move = lambda: None


patch = _patch
unpatch = _unpatch
move = _move



"""
"""
from __future__ import annotations

import multiprocessing
import os
import signal
import traceback
import warnings
from concurrent.futures import ThreadPoolExecutor
from contextlib import nullcontext
from contextvars import copy_context
from datetime import timedelta
from functools import partial
from functools import wraps
from multiprocessing.context import ForkProcess
from pickle import PicklingError
from queue import Empty
from queue import Queue as ThreadQueue
from threading import Thread
from typing import TYPE_CHECKING
from typing import Callable
from typing import Generator
from typing import Generic
from typing_extensions import assert_never

import psutil

from ..config import Config
from ..reloading import reload_server_ports
from ..utils import debug
from ..utils import drop_params
from ..utils import gradio_request_var
from ..utils import SimpleQueue as Queue
from . import client
from . import torch
from .api import AllowToken
from .api import NvidiaIndex
from .api import NvidiaUUID
from .gradio import GradioPartialContext
from .gradio import error
from .gradio import get_server_port
from .gradio import patch_gradio_queue
from .gradio import try_process_queue_event
from .tqdm import remove_tqdm_multiprocessing_lock
from .tqdm import tqdm
from .utils import apply_cleanups
from .types import * # TODO: Please don't do that


GENERATOR_GLOBAL_TIMEOUT = 20 * 60

SPAWN_PROGRESS_CLEANUP = 0.1
SPAWN_PROGRESS_INIT = 0.1


Process = multiprocessing.get_context('fork').Process
forked = False


class Worker(Generic[Res]):
    process: ForkProcess
    arg_queue: Queue[tuple[Params, GradioPartialContext]]
    res_queue: Queue[Res | None]
    _sentinel: Thread

    def __init__(
        self,
        target: Callable[[
            Queue[tuple[Params, GradioPartialContext]],
            Queue[Res | None],
            AllowToken,
            NvidiaUUID,
            list[int],
        ], None],
        allow_token: str,
        nvidia_uuid: str,
    ):
        self._sentinel = Thread(target=self._close_on_exit, daemon=True)
        self.arg_queue = Queue()
        self.res_queue = Queue()
        debug(f"{self.arg_queue._writer.fileno()=}") # pyright: ignore [reportAttributeAccessIssue]
        debug(f"{self.res_queue._writer.fileno()=}") # pyright: ignore [reportAttributeAccessIssue]
        server_ports: set[int] = set()
        if (gradio_server_port := get_server_port()) is not None:
            server_ports |= {gradio_server_port}
        server_ports |= reload_server_ports
        fds = [conn.fd for conn in psutil.Process().connections() if conn.laddr.port in server_ports]
        args = self.arg_queue, self.res_queue, allow_token, nvidia_uuid, fds
        if TYPE_CHECKING:
            target(*args)
        self.process = Process(
            target=target,
            args=args,
            daemon=True,
        )
        self.process.start()
        self._sentinel.start()

    def _close_on_exit(self):
        assert (pid := self.process.pid) is not None
        self.process.join()
        self.arg_queue.close()
        self.res_queue.wlock_release()
        apply_cleanups(pid)
        self.res_queue.put(None)


def worker_init(
    res_queue: Queue[RegularResQueueResult | None] | Queue[GeneratorResQueueResult | None],
    allow_token: str,
    nvidia_uuid: str,
    fds: list[int],
) -> None | ExceptionResult:
    # Immediately close file descriptors
    for fd in fds:
        try:
            os.close(fd)
        except Exception as e: # pragma: no cover
            if isinstance(e, OSError) and e.errno == 9:
                continue
            return exception_result(e)
    try:
        remove_tqdm_multiprocessing_lock()
    except Exception: # pragma: no cover
        print("Error while trying to remove tqdm mp_lock:")
        traceback.print_exc()
    progress = nullcontext()
    if tqdm is not None and Config.zero_gpu_v2:
        progress = tqdm(total=100, desc="ZeroGPU init", file=open(os.devnull, 'w'))
    try: # Unrecoverable init part
        patch_gradio_queue(res_queue)
        with progress as progress:
            current_progress = 0 # Gradio does not support float progress updates
            def update(n: float):
                nonlocal current_progress
                current_progress += n
                if progress is not None:
                    progress.update(round(current_progress * 100) - progress.n)
            client.allow(allow_token)
            update(SPAWN_PROGRESS_CLEANUP)
            torch.unpatch()
            torch.init(nvidia_uuid)
            update(SPAWN_PROGRESS_INIT)
            callback = None
            if (transfer_size := torch.size()) > 0:
                remaining = 1 - (SPAWN_PROGRESS_CLEANUP + SPAWN_PROGRESS_INIT)
                callback = lambda n: update(n * remaining / transfer_size)
            torch.move(callback=callback)
    except Exception as e: # pragma: no cover
        return exception_result(e)


def process_duration(duration: Duration | None):
    if duration is None or isinstance(duration, timedelta):
        return duration
    return timedelta(seconds=duration)


def static_duration(duration: DynamicDuration[Param], *args: Param.args, **kwargs: Param.kwargs):
    if not callable(duration):
        return duration
    return duration(*args, **kwargs)


def regular_function_wrapper(
    task: Callable[Param, Res],
    duration: DynamicDuration[Param],
) -> Callable[Param, Res]:

    import gradio as gr

    request_var = gradio_request_var()
    workers: dict[NvidiaIndex, Worker[RegularResQueueResult[Res] | None]] = {}
    task_id = id(task)

    @wraps(task)
    def gradio_handler(*args: Param.args, **kwargs: Param.kwargs) -> Res:

        if forked:
            return task(*args, **kwargs)

        request = request_var.get(None)
        duration_ = static_duration(duration, *args, **kwargs)
        duration_ = process_duration(duration_)
        schedule_response = client.schedule(task_id=task_id, request=request, duration=duration_)
        allow_token = schedule_response.allowToken
        nvidia_index = schedule_response.nvidiaIndex
        nvidia_uuid = schedule_response.nvidiaUUID
        release = partial(client.release, allow_token)

        try:
            worker = workers.pop(nvidia_index)
        except KeyError:
            worker = None

        if worker is not None and worker.process.is_alive() and schedule_response.idle:
            assert worker.arg_queue.empty()
            assert worker.res_queue.empty()
        else:
            worker = Worker(thread_wrapper, allow_token, nvidia_uuid)

        try:
            worker.arg_queue.put(((args, kwargs), GradioPartialContext.get()))
        except PicklingError: # TODO: detailed serialization diagnostic
            release(fail=True)
            raise

        while True:
            res = worker.res_queue.get()
            if res is None:
                release(fail=True, allow_404=True)
                raise error("ZeroGPU worker error", "GPU task aborted")
            if isinstance(res, ExceptionResult):
                release(fail=True)
                print(res.traceback)
                raise error("ZeroGPU worker error", res.error_cls)
            if isinstance(res, OkResult):
                workers[nvidia_index] = worker
                release()
                return res.value
            if isinstance(res, GradioQueueEvent):
                try_process_queue_event(res.method_name, *res.args, **res.kwargs)
                continue
            assert_never(res)


    def thread_wrapper(
        arg_queue: Queue[tuple[Params, GradioPartialContext]],
        res_queue: Queue[RegularResQueueResult[Res] | None],
        allow_token: str,
        nvidia_uuid: str,
        fds: list[int],
    ):
        global forked
        forked = True
        signal.signal(signal.SIGTERM, drop_params(arg_queue.close))
        initialized = False
        while True:
            try:
                (args, kwargs), gradio_context = arg_queue.get()
            except OSError:
                break
            if not initialized:
                if (res := worker_init(
                    res_queue=res_queue,
                    allow_token=allow_token,
                    nvidia_uuid=nvidia_uuid,
                    fds=fds,
                )) is not None:
                    res_queue.put(res)
                    return
                initialized = True
            def run_task():
                try:
                    res = task(*args, **kwargs) # pyright: ignore [reportCallIssue]
                except Exception as e:
                    res = exception_result(e)
                else:
                    res = OkResult(res)
                try:
                    res_queue.put(res)
                except PicklingError as e:
                    res_queue.put(exception_result(e))
            GradioPartialContext.apply(gradio_context)
            context = copy_context()
            with ThreadPoolExecutor() as executor:
                future = executor.submit(context.run, run_task)
            future.result()

    # https://github.com/python/cpython/issues/91002
    if not hasattr(task, '__annotations__'):
        gradio_handler.__annotations__ = {}

    return gradio_handler


def generator_function_wrapper(
    task: Callable[Param, Generator[Res, None, None]],
    duration: DynamicDuration[Param],
) -> Callable[Param, Generator[Res, None, None]]:

    import gradio as gr

    request_var = gradio_request_var()
    workers: dict[NvidiaIndex, Worker[GeneratorResQueueResult[Res] | None]] = {}
    task_id = id(task)

    @wraps(task)
    def gradio_handler(*args: Param.args, **kwargs: Param.kwargs) -> Generator[Res, None, None]:

        if forked:
            yield from task(*args, **kwargs)
            return

        request = request_var.get(None)
        duration_ = static_duration(duration, *args, **kwargs)
        duration_ = process_duration(duration_)
        schedule_response = client.schedule(task_id=task_id, request=request, duration=duration_)
        allow_token = schedule_response.allowToken
        nvidia_index = schedule_response.nvidiaIndex
        nvidia_uuid = schedule_response.nvidiaUUID
        release = partial(client.release, allow_token)

        try:
            worker = workers.pop(nvidia_index)
        except KeyError:
            worker = None

        if worker is not None and worker.process.is_alive() and schedule_response.idle:
            assert worker.arg_queue.empty()
            assert worker.res_queue.empty()
        else:
            worker = Worker(thread_wrapper, allow_token, nvidia_uuid)

        try:
            worker.arg_queue.put(((args, kwargs), GradioPartialContext.get()))
        except PicklingError: # TODO: detailed serialization diagnostic
            release(fail=True)
            raise

        yield_queue: ThreadQueue[YieldQueueResult[Res]] = ThreadQueue()
        def fill_yield_queue(worker: Worker[GeneratorResQueueResult[Res] | None]):
            while True:
                res = worker.res_queue.get()
                if res is None:
                    release(fail=True, allow_404=True)
                    yield_queue.put(AbortedResult())
                    return
                if isinstance(res, ExceptionResult):
                    release(fail=True)
                    yield_queue.put(res)
                    return
                if isinstance(res, EndResult):
                    workers[nvidia_index] = worker
                    release()
                    yield_queue.put(EndResult())
                    return
                if isinstance(res, OkResult):
                    yield_queue.put(OkResult(res.value))
                    continue
                if isinstance(res, GradioQueueEvent): # pragma: no cover (not working properly on Gradio side)
                    try_process_queue_event(res.method_name, *res.args, **res.kwargs)
                    continue
                debug(f"fill_yield_queue: assert_never({res=})")
                assert_never(res)
        from typing_extensions import assert_never
        with ThreadPoolExecutor() as e:
            f = e.submit(copy_context().run, fill_yield_queue, worker)
            f.add_done_callback(lambda _: debug("fill_yield_queue DONE"))
            while True:
                try:
                    res = yield_queue.get(timeout=GENERATOR_GLOBAL_TIMEOUT)
                except Empty: # pragma: no cover
                    debug(f"yield_queue TIMEOUT ({GENERATOR_GLOBAL_TIMEOUT=})")
                    raise
                if isinstance(res, AbortedResult):
                    raise error("ZeroGPU worker error", "GPU task aborted")
                if isinstance(res, ExceptionResult):
                    print(res.traceback)
                    raise error("ZeroGPU worker error", res.error_cls)
                if isinstance(res, EndResult):
                    break
                if isinstance(res, OkResult):
                    yield res.value
                    continue
                debug(f"gradio_handler: assert_never({res=})")
                assert_never(res)


    def thread_wrapper(
        arg_queue: Queue[tuple[Params, GradioPartialContext]],
        res_queue: Queue[GeneratorResQueueResult[Res] | None],
        allow_token: str,
        nvidia_uuid: str,
        fds: list[int],
    ):
        global forked
        forked = True
        signal.signal(signal.SIGTERM, drop_params(arg_queue.close))
        initialized = False
        while True:
            try:
                (args, kwargs), gradio_context = arg_queue.get()
            except OSError:
                break
            if not initialized:
                if (res := worker_init(
                    res_queue=res_queue,
                    allow_token=allow_token,
                    nvidia_uuid=nvidia_uuid,
                    fds=fds,
                )) is not None:
                    res_queue.put(res)
                    return
                initialized = True
            def iterate():
                gen = task(*args, **kwargs) # type: ignore
                while True:
                    try:
                        res = next(gen)
                    except StopIteration:
                        break
                    except Exception as e:
                        res_queue.put(exception_result(e))
                        break
                    try:
                        res_queue.put(OkResult(res))
                    except PicklingError as e:
                        res_queue.put(exception_result(e))
                        break
                    else:
                        continue
            GradioPartialContext.apply(gradio_context)
            with ThreadPoolExecutor() as executor:
                executor.submit(copy_context().run, iterate)
            res_queue.put(EndResult())

    # https://github.com/python/cpython/issues/91002
    if not hasattr(task, '__annotations__'):
        gradio_handler.__annotations__ = {}

    return gradio_handler


def exception_result(exc: Exception) -> ExceptionResult:
    formatted = traceback.format_exception(type(exc), exc, exc.__traceback__)
    return ExceptionResult(traceback=''.join(formatted), error_cls=exc.__class__.__name__)



"""
"""
from __future__ import annotations

from typing import NamedTuple

import torch


class AliasId(NamedTuple):
    data_ptr: int
    dtype: torch.dtype
    shape: tuple[int, ...]
    stride: tuple[int, ...]

    @classmethod
    def from_tensor(cls, tensor: torch.Tensor):
        return cls(
            tensor.data_ptr(),
            tensor.dtype,
            tensor.shape,
            tensor.stride(),
        )



"""
"""

from types import SimpleNamespace as _SimpleNamespace

import torch as _torch

from ...config import Config


def compute_base_free_memory(total_memory: int):
    pytorch_base_memory = 309002240 # TODO: fine-grain per: torch-version x GPU(-MIG) model
    return total_memory - pytorch_base_memory - Config.zerogpu_cuda_reserved_memory

CUDA_DEVICE_NAME = Config.zerogpu_cuda_device_name
CUDA_TOTAL_MEMORY = Config.zerogpu_cuda_total_memory
CUDA_MEM_GET_INFO = (compute_base_free_memory(CUDA_TOTAL_MEMORY), CUDA_TOTAL_MEMORY)
CUDA_DEVICE_CAPABILITY = (Config.zerogpu_cuda_capability_major, Config.zerogpu_cuda_capability_minor)
CUDA_DEVICE_PROPERTIES = _SimpleNamespace(
    name=CUDA_DEVICE_NAME,
    major=CUDA_DEVICE_CAPABILITY[0],
    minor=CUDA_DEVICE_CAPABILITY[1],
    total_memory=CUDA_TOTAL_MEMORY,
    multi_processor_count=Config.zerogpu_cuda_multi_processor_count,
    # TODO: L2_cache_size
)

if _torch.version.cuda.startswith("12."): # pyright: ignore [reportAttributeAccessIssue, reportOptionalMemberAccess]
    CUDA_MEMORY_STATS_AS_NESTED_DICT = {
        "num_alloc_retries": 0,
        "num_ooms": 0,
        "max_split_size": -1,
        "num_sync_all_streams": 0,
        "num_device_alloc": 0,
        "num_device_free": 0,
        "allocation": {
            "all": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "small_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "large_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
        },
        "segment": {
            "all": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "small_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "large_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
        },
        "active": {
            "all": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "small_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "large_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
        },
        "inactive_split": {
            "all": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "small_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "large_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
        },
        "allocated_bytes": {
            "all": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "small_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "large_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
        },
        "reserved_bytes": {
            "all": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "small_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "large_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
        },
        "active_bytes": {
            "all": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "small_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "large_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
        },
        "inactive_split_bytes": {
            "all": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "small_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "large_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
        },
        "requested_bytes": {
            "all": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "small_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "large_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
        },
        "oversize_allocations": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
        "oversize_segments": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
    }
else: # pragma: no cover (CUDA 11)
    CUDA_MEMORY_STATS_AS_NESTED_DICT = {
        "num_alloc_retries": 0,
        "num_ooms": 0,
        "max_split_size": -1,
        "allocation": {
            "all": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "small_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "large_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
        },
        "segment": {
            "all": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "small_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "large_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
        },
        "active": {
            "all": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "small_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "large_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
        },
        "inactive_split": {
            "all": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "small_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "large_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
        },
        "allocated_bytes": {
            "all": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "small_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "large_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
        },
        "reserved_bytes": {
            "all": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "small_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "large_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
        },
        "active_bytes": {
            "all": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "small_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "large_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
        },
        "inactive_split_bytes": {
            "all": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "small_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "large_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
        },
        "requested_bytes": {
            "all": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "small_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
            "large_pool": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
        },
        "oversize_allocations": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
        "oversize_segments": {"current": 0, "peak": 0, "allocated": 0, "freed": 0},
    }



import gc
import importlib
import sys
import time
import types
from collections import defaultdict
from inspect import getattr_static

MAX_TIME = 0.1


def _make_audithook(self):
    def watch_exec(event, obj):  # pragma: no cover
        # Note: Python does not trace audit hooks, so normal use will not show
        # coverage of this function even if it is executed
        if event == "exec":
            (code,) = obj
            self.assimilate(code)

    # The closure is faster than a method on _CodeDB
    return watch_exec


class ConformException(Exception):
    pass


class CodeRegistry:
    def __init__(self):
        self.codes = {}
        self.currcodes = {}
        self.backcodes = defaultdict(set)
        self.functions = defaultdict(set)
        self.last_cost = 0
        self.always_use_cache = False

    def setup(self):
        self.collect_all()
        sys.addaudithook(_make_audithook(self))

    def collect_all(self):
        # Collect code objects
        results = []
        for obj in gc.get_objects():
            try:
                if isinstance(obj, types.FunctionType):
                    results.append((obj, obj.__code__))
                elif getattr_static(obj, "__conform__", None) is not None:
                    for x in gc.get_referents(obj):
                        if isinstance(x, types.CodeType):
                            results.append((obj, x))
            except ReferenceError:  # pragma: no cover
                # Object got deleted
                pass
        for obj, co in results:
            if isinstance((qual := getattr(obj, "__qualname__", None)), str):
                qualpath = [
                    part for part in qual.split(".")[:-1] if part != "<locals>"
                ]
                self.assimilate(co, (co.co_filename, *qualpath))
            self.functions[co].add(obj)

    def _setcodepaths(self, paths, code):
        for p in paths:
            self.codes.setdefault(p, code)
            self.currcodes[p] = code
        self.backcodes[code].update(paths)

    def assimilate(self, code, path=()):
        if code.co_name == "<module>":  # pragma: no cover
            # TODO: Not tested and I forget what the problem is.
            # I had to add a condition, because it seems that it would get
            # triggered in other situations.
            if not path or path[-1] != code.co_filename:
                # Typically triggered by the audit hook
                name = code.co_filename
            else:
                name = None
        elif code.co_name.startswith(
            "<generic parameters of"
        ):  # pragma: no cover
            # We'll loop through the co_consts which contains the actual class
            name = None
        elif code.co_name.startswith("<"):
            return
        else:
            name = code.co_name
        if name:
            path = (*path, name)
            self._setcodepaths(
                [(*path, code.co_firstlineno), (*path, None)], code
            )
        for ct in code.co_consts:
            if isinstance(ct, types.CodeType):
                self.assimilate(ct, path)

    def _get_functions(self, code):
        t = time.time()
        results = [
            fn
            for fn in gc.get_referrers(code)
            if isinstance(fn, types.FunctionType) or hasattr(fn, "__conform__")
        ]
        self.functions[code] = set(results)
        self.last_cost = time.time() - t
        return results

    def get_functions(self, code, use_cache=False):
        use_cache = (
            use_cache or self.always_use_cache or self.last_cost > MAX_TIME
        )
        if use_cache and (results := self.functions[code]):
            return list(results)
        else:
            return self._get_functions(code)

    def update_cache_entry(self, obj, old_code, new_code):
        self.functions[old_code].discard(obj)
        self.functions[new_code].add(obj)
        self._setcodepaths(self.backcodes[old_code], new_code)

    def find_code(self, *path, filename=None, module=None, lineno=None):
        if module is not None:
            assert filename is None
            filename = importlib.import_module(module).__file__
        assert filename is not None
        path = (filename, *path, lineno)
        return self.currcodes[path]

    def conform(self, obj1, obj2, use_cache=False):
        if hasattr(obj1, "__conform__"):
            obj1.__conform__(obj2)

        elif isinstance(obj1, types.CodeType):
            for fn in self.get_functions(obj1, use_cache=use_cache):
                self.conform(fn, obj2)

        elif isinstance(obj2, types.FunctionType):
            self.conform(obj1, obj2.__code__)
            obj1.__defaults__ = obj2.__defaults__
            obj1.__kwdefaults__ = obj2.__kwdefaults__

        elif isinstance(obj2, types.CodeType):
            fv1 = obj1.__code__.co_freevars
            fv2 = obj2.co_freevars
            if fv1 != fv2:
                msg = (
                    f"Cannot replace closure `{obj1.__name__}` because the free "
                    f"variables changed. Before: {fv1}; after: {fv2}."
                )
                if ("__class__" in (fv1 or ())) ^ ("__class__" in (fv2 or ())):
                    msg += " Note: The use of `super` entails the `__class__` free variable."
                raise ConformException(msg)
            self.update_cache_entry(obj1, obj1.__code__, obj2)
            obj1.__code__ = obj2

        elif obj2 is None:
            pass

        else:  # pragma: no cover
            raise ConformException(f"Cannot conform {obj1} with {obj2}")



"""
"""
# pyright: reportPrivateImportUsage=false

from __future__ import annotations

from importlib import metadata

from packaging import version


MULTI_BACKEND_VERSION = version.parse('0.46.0')


def is_old_bnb() -> bool:
    try:
        version_str = metadata.version('bitsandbytes')
    except ImportError: # pragma: no cover
        return False
    if (bnb_version := version.parse(version_str)) < version.parse('0.46.0'): # pragma: no cover
        message = f"ZeroGPU recommends bitsandbytes >= `{MULTI_BACKEND_VERSION}` "
        message += f"(`{bnb_version}` installed). Falling back to legacy support"
        print(message)
        return True
    return False


if is_old_bnb(): # pragma: no cover

    from . import bitsandbytes_legacy

    _patch = bitsandbytes_legacy.patch
    _unpatch = bitsandbytes_legacy.unpatch
    _move = bitsandbytes_legacy.move

else:

    _patch = lambda: None
    _unpatch = lambda: None
    _move = lambda: None


patch = _patch
unpatch = _unpatch
move = _move



"""
"""

from multiprocessing.synchronize import RLock as MultiprocessingRLock


try:
    from tqdm import tqdm as _tqdm
except ImportError: # pragma: no cover
    _tqdm = None


def remove_tqdm_multiprocessing_lock():
    if _tqdm is None: # pragma: no cover
        return
    tqdm_lock = _tqdm.get_lock()
    assert tqdm_lock.__class__.__name__ == 'TqdmDefaultWriteLock'
    tqdm_lock.locks = [
        lock for lock in tqdm_lock.locks
        if not isinstance(lock, MultiprocessingRLock)
    ]


tqdm = _tqdm



"""
"""
from __future__ import annotations

import sys
from functools import partial

import multiprocessing
from multiprocessing.queues import SimpleQueue as _SimpleQueue
from pickle import PicklingError
from threading import Thread
from typing import Any
from typing import Callable
from typing import TypeVar

from typing_extensions import ParamSpec


GRADIO_VERSION_ERROR_MESSAGE = "Make sure Gradio version is at least 3.46"


T = TypeVar('T')
P = ParamSpec('P')


if sys.version_info.minor < 9: # pragma: no cover
    _SimpleQueue.__class_getitem__ = classmethod(lambda cls, _: cls) # type: ignore

class SimpleQueue(_SimpleQueue[T]):
    def __init__(self, *args):
        super().__init__(*args, ctx=multiprocessing.get_context('fork'))
    def put(self, obj: T):
        try:
            super().put(obj)
        except PicklingError:
            raise # pragma: no cover
        # https://bugs.python.org/issue29187
        except Exception as e:
            message = str(e)
            if not "pickle" in message:
                raise # pragma: no cover
            raise PicklingError(message)
    def close(self): # Python 3.8 static typing trick
        super().close() # type: ignore
    def wlock_release(self):
        if (lock := getattr(self, '_wlock', None)) is None:
            return # pragma: no cover
        try:
            lock.release()
        except ValueError:
            pass


def drop_params(fn: Callable[[], T]) -> Callable[..., T]:
    def drop(*args):
        return fn()
    return drop


def gradio_request_var():
    try:
        from gradio.context import LocalContext
    except ImportError: # pragma: no cover
        raise RuntimeError(GRADIO_VERSION_ERROR_MESSAGE)
    return LocalContext.request


debug = partial(print, 'SPACES_ZERO_GPU_DEBUG')


# Type-safe threads
def create_thread(fn: Callable[P, Any], *args: P.args, **kwargs: P.kwargs) -> Thread:
    return Thread(target=fn, args=args, kwargs=kwargs)



"""
"""
from __future__ import annotations

import base64
import ctypes
import json
import shutil
from functools import cache
from pathlib import Path
from uuid import uuid4
from typing import Any

from ..config import ZEROGPU_HOME
from ..config import Config


CLEANUPS_BASE_DIR = ZEROGPU_HOME / 'cleanups'


def register_cleanup(pid: int, target_dir: Path):
    cleanups_dir = CLEANUPS_BASE_DIR / f'{pid}'
    cleanups_dir.mkdir(parents=True, exist_ok=True)
    cleanup = cleanups_dir / f'{uuid4()}'
    cleanup.symlink_to(target_dir, target_is_directory=True)


def apply_cleanups(pid: int):
    cleanups_dir = CLEANUPS_BASE_DIR / f'{pid}'
    try:
        targets = [cleanup.readlink() for cleanup in cleanups_dir.iterdir()]
    except FileNotFoundError:
        return
    for target in targets:
        shutil.rmtree(target, ignore_errors=True)
    shutil.rmtree(cleanups_dir, ignore_errors=True)


@cache
def self_cgroup_device_path() -> str:
    cgroup_content = Path(Config.zerogpu_proc_self_cgroup_path).read_text()
    for line in cgroup_content.strip().split('\n'):
        contents = line.split(':devices:')
        if len(contents) != 2:
            continue # pragma: no cover
        return contents[1]
    raise Exception # pragma: no cover


def malloc_trim():
    ctypes.CDLL("libc.so.6").malloc_trim(0)


def jwt_payload(token: str) -> dict[str, Any]:
    _, payload, _ = token.split('.')
    return json.loads(base64.urlsafe_b64decode(f'{payload}=='))



import ast
from dataclasses import dataclass, field, replace as dc_replace

from ovld import ovld, recurse


@dataclass
class Variables:
    assigned: set = field(default_factory=set)
    read: set = field(default_factory=set)
    closure: set = field(default_factory=set)

    @property
    def free(self):
        return self.read - self.assigned

    replace = dc_replace

    def __or__(self, other):
        return Variables(
            assigned=self.assigned | other.assigned,
            read=self.read | other.read,
        )


@ovld
def variables(seq: list, mapping):
    fvs = Variables()
    for node in seq:
        fvs = fvs | recurse(node, mapping)
    return fvs


@ovld
def variables(node: (ast.FunctionDef, ast.AsyncFunctionDef), mapping):
    fvs = (
        recurse(node.body, mapping)
        | recurse(node.args.args, mapping)
        | recurse(node.args.posonlyargs, mapping)
        | recurse(node.args.kwonlyargs, mapping)
        | recurse(node.args.kwarg, mapping)
        | recurse(node.args.vararg, mapping)
    )
    mapping[node] = fvs
    outer = (
        recurse(node.decorator_list, mapping)
        | recurse(node.args.defaults, mapping)
        | recurse(node.args.kw_defaults, mapping)
    )
    return outer | Variables(assigned={node.name}, read=fvs.free)


@ovld
def variables(node: ast.ClassDef, mapping):
    fvs = recurse(node.body, mapping) | Variables(assigned={"__class__"})
    mapping[node] = fvs
    outer = recurse(node.decorator_list, mapping)
    return outer | Variables(assigned={node.name}, read=fvs.free)


@ovld
def variables(node: ast.arg, mapping):
    return Variables(assigned={node.arg})


@ovld
def variables(node: ast.Name, mapping):
    if isinstance(node.ctx, ast.Load):
        read = {node.id}
        if node.id == "super":
            read.add("__class__")
        return Variables(read=read)
    elif isinstance(node.ctx, ast.Store):
        return Variables(assigned={node.id})
    else:
        return Variables(read={node.id})


@ovld
def variables(node: ast.AST, mapping):
    return recurse(list(ast.iter_child_nodes(node)), mapping)


@ovld  # pragma: no cover
def variables(thing: object, mapping):
    # Just in case
    return Variables()



"""
"""
from __future__ import annotations

from typing import Literal

from pydantic import BaseModel


class ReloadRegion(BaseModel):
    startLine: int
    startCol: int
    endLine: int
    endCol: int


class ReloadOperationObject(BaseModel):
    kind: Literal['add', 'update', 'delete']
    region: ReloadRegion
    objectType: str
    objectName: str


class ReloadOperationRun(BaseModel):
    kind: Literal['run']
    region: ReloadRegion
    codeLines: str
    stdout: str | None = None
    stderr: str | None = None


class ReloadOperationException(BaseModel):
    kind: Literal['exception']
    region: ReloadRegion
    traceback: str


class ReloadOperationError(BaseModel):
    kind: Literal['error']
    traceback: str


class ReloadOperationUI(BaseModel):
    kind: Literal['ui']
    updated: bool


class ApiCreateReloadRequest(BaseModel):
    filepath: str
    contents: str


class ApiCreateReloadResponseSuccess(BaseModel):
    status: Literal['created']
    reloadId: str


class ApiCreateReloadResponseError(BaseModel):
    status: Literal['alreadyReloading', 'fileNotFound']


class ApiCreateReloadResponse(BaseModel):
    res: ApiCreateReloadResponseError | ApiCreateReloadResponseSuccess


class ApiGetReloadRequest(BaseModel):
    reloadId: str


class ApiGetReloadEventSourceData(BaseModel):
    data: ReloadOperationError \
        | ReloadOperationException \
        | ReloadOperationObject \
        | ReloadOperationRun \
        | ReloadOperationUI \


class ApiGetStatusRequest(BaseModel):
    revision: str


class ApiGetStatusResponse(BaseModel):
    reloading: bool
    uncommited: list[str]


class ApiFetchContentsRequest(BaseModel):
    filepath: str


class ApiFetchContentsResponseError(BaseModel):
    status: Literal['fileNotFound']


class ApiFetchContentsResponseSuccess(BaseModel):
    status: Literal['ok']
    contents: str


class ApiFetchContentsResponse(BaseModel):
    res: ApiFetchContentsResponseError | ApiFetchContentsResponseSuccess



"""
"""
from __future__ import annotations

import time

import ctypes
import os
from concurrent.futures import as_completed
from concurrent.futures import ThreadPoolExecutor
from contextvars import copy_context
from dataclasses import dataclass
from queue import Queue
from typing import Callable

from ...utils import debug
from .utils import empty_like_raw_alloc

import torch
from typing_extensions import TypeAlias


PAGE_SIZE = 4096
TOTAL_MEMORY = os.sysconf('SC_PAGE_SIZE') * os.sysconf('SC_PHYS_PAGES')
VM_MAX_SIZE = min(2**38, TOTAL_MEMORY // 2)

BUFFER_SIZE = 128 * 2**20
BUFFER_COUNT = 2


TensorWithSizes: TypeAlias = 'tuple[torch.Tensor, int, int]'

@dataclass
class ZeroGPUTensorPack:
    base_dir: str
    batches: list[list[TensorWithSizes]]
    big_tensors: list[TensorWithSizes]
    fakes: dict[torch.Tensor, list[torch.Tensor]]
    total_size: int
    def path(self):
        return f'{self.base_dir}/{id(self)}'
    def __del__(self):
        try:
            os.remove(self.path())
        except FileNotFoundError: # pragma: no cover
            pass


def write(fd: int, tensor: torch.Tensor):
    clone = torch.empty_like(tensor)
    size = clone.untyped_storage().size() # pyright: ignore [reportAttributeAccessIssue]
    buffer = torch.UntypedStorage(VM_MAX_SIZE)
    buffer_ptr = buffer.data_ptr()
    offset = -buffer_ptr % PAGE_SIZE
    padding = -size % PAGE_SIZE
    clone.set_(buffer[offset:offset+size], 0, clone.shape, clone.stride()) # pyright: ignore [reportArgumentType]
    clone.copy_(tensor)
    mv = memoryview((ctypes.c_char * (size+padding)).from_address(buffer_ptr+offset))
    written_bytes = 0
    while written_bytes < size:
        written_bytes += os.write(fd, mv[written_bytes:])


def pack_tensors(
    tensors: set[torch.Tensor],
    fakes: dict[torch.Tensor, list[torch.Tensor]],
    offload_dir: str,
    callback: Callable[[int]] | None = None,
):

    callback = (lambda bytes: None) if callback is None else callback

    batches: list[list[TensorWithSizes]] = []
    big_tensors: list[TensorWithSizes] = []

    tensors_with_sizes: list[tuple[torch.Tensor, int, int]] = []
    for tensor in tensors:
        size = tensor.numel() * tensor.element_size()
        aligned_size = size + (-size % PAGE_SIZE)
        tensors_with_sizes += [(tensor, size, aligned_size)]

    current_batch, current_size = [], 0
    for (tensor, size, aligned_size) in sorted(tensors_with_sizes, key=lambda item: item[2]):
        if aligned_size > BUFFER_SIZE:
            big_tensors += [(tensor, size, aligned_size)]
            continue
        current_size += aligned_size
        if current_size > BUFFER_SIZE:
            batches += [current_batch]
            current_batch, current_size = [(tensor, size, aligned_size)], aligned_size
        else:
            current_batch += [(tensor, size, aligned_size)]

    if current_batch:
        batches += [current_batch]

    get_meta = {tensor: empty_like_raw_alloc(tensor) for tensor in tensors}
    batches_meta = [[(get_meta[tensor], size, asize) for tensor, size, asize in batch] for batch in batches]
    big_tensors_meta = [(get_meta[tensor], size, asize) for tensor, size, asize in big_tensors]
    fakes_meta = {get_meta[tensor]: fake_list for tensor, fake_list in fakes.items()}

    pack = ZeroGPUTensorPack(
        base_dir=offload_dir,
        batches=batches_meta,
        big_tensors=big_tensors_meta,
        fakes=fakes_meta,
        total_size=sum([size for _, size, _ in tensors_with_sizes]),
    )

    fd = os.open(pack.path(), os.O_CREAT | os.O_WRONLY | os.O_DIRECT)
    try:
        total_asize = sum([aligned_size for batch in batches for *_, aligned_size in batch])
        total_asize += sum([aligned_size for *_, aligned_size in big_tensors])
        if total_asize > 0:
            os.posix_fallocate(fd, 0, total_asize)
            for batch in batches:
                for tensor, size, _ in batch:
                    write(fd, tensor)
                    callback(size)
            for tensor, size, _ in big_tensors:
                write(fd, tensor)
                callback(size)
        return pack
    finally:
        os.close(fd)


def pack_to_cuda(pack: ZeroGPUTensorPack, callback: Callable[[int]] | None = None):

    callback = (lambda bytes: None) if callback is None else callback

    free_buffers: Queue[torch.Tensor] = Queue()
    read_buffers: Queue[torch.Tensor] = Queue()

    for _ in range(BUFFER_COUNT):
        free_buffers.put(torch.ByteTensor(BUFFER_SIZE).pin_memory())

    def read(fd: int, buffer: torch.Tensor, size: int):
        mv = memoryview((ctypes.c_char * size).from_address(buffer.data_ptr()))
        read_bytes = 0
        while read_bytes < size:
            read_bytes += os.readv(fd, [mv[read_bytes:]])

    def disk_to_pin(fd: int):
        for batch in pack.batches:
            buffer = free_buffers.get()
            batch_size = sum([aligned_size for *_, aligned_size in batch])
            read(fd, buffer, batch_size)
            read_buffers.put(buffer)
        for *_, aligned_size in pack.big_tensors:
            read_bytes = 0
            while read_bytes < aligned_size:
                buffer = free_buffers.get()
                read_size = min(BUFFER_SIZE, aligned_size - read_bytes)
                read(fd, buffer, read_size)
                read_buffers.put(buffer)
                read_bytes += read_size

    def pin_to_cuda():
        total_duration_in_callback = 0
        for batch in pack.batches:
            buffer = read_buffers.get()
            offset = 0
            cuda_storages = []
            for tensor, size, aligned_size in batch:
                cuda_storages += [buffer[offset:offset+size].cuda(non_blocking=True)]
                offset += aligned_size
            torch.cuda.synchronize()
            free_buffers.put(buffer)
            batch_total_size = 0
            for (tensor, size, _), cuda_storage in zip(batch, cuda_storages):
                cuda_tensor = torch.tensor([], dtype=tensor.dtype, device='cuda')
                cuda_tensor = cuda_tensor.set_(cuda_storage.untyped_storage(), 0, tensor.shape, tensor.stride())
                for fake in pack.fakes[tensor]:
                    fake.data = cuda_tensor
                batch_total_size += size
            t0 = time.perf_counter()
            callback(batch_total_size)
            total_duration_in_callback += time.perf_counter() - t0
        for tensor, size, _ in pack.big_tensors:
            cuda_storage = torch.empty(size, dtype=torch.uint8, device='cuda')
            offset = 0
            while offset < size:
                buffer = read_buffers.get()
                read_size = min(BUFFER_SIZE, size - offset)
                cuda_storage[offset:offset+read_size] = buffer[:read_size]
                offset += read_size
                torch.cuda.synchronize() # Probably not needed
                free_buffers.put(buffer)
                t0 = time.perf_counter()
                callback(read_size)
                total_duration_in_callback += time.perf_counter() - t0
            cuda_tensor = torch.tensor([], dtype=tensor.dtype, device='cuda')
            cuda_tensor = cuda_tensor.set_(cuda_storage.untyped_storage(), 0, tensor.shape, tensor.stride())
            for fake in pack.fakes[tensor]:
                fake.data = cuda_tensor

        debug(f"{total_duration_in_callback=}")

    with ThreadPoolExecutor(2) as e:
        fd = os.open(pack.path(), os.O_RDONLY | os.O_DIRECT)
        try:
            futures = [
                e.submit(copy_context().run, disk_to_pin, fd),
                e.submit(copy_context().run, pin_to_cuda),
            ]
            for future in as_completed(futures):
                future.result()
        finally:
            os.close(fd)



"""
Synced with huggingface/pyspaces:spaces/zero/api.py
"""
from __future__ import annotations

from datetime import timedelta
from typing import Any
from typing import Generator
from typing import Literal
from typing import NamedTuple
from typing import Optional
from typing import overload

import httpx
from pydantic import BaseModel
from typing_extensions import assert_never


AllowToken = str
NvidiaIndex = int # TODO: Migrate to GpuIndex (less confusing for MIG)
NvidiaUUID = str
CGroupPath = str
TaskId = int

GPUSize = Literal['medium', 'large']
AuthLevel = Literal['regular', 'pro']
QueuingReason = Literal['node', 'concurrency']


AUTHENTICATED_HEADER = 'X-Authenticated'
QUEUING_REASON_HEADER = 'X-Queuing-Reason'


class ScheduleResponse(BaseModel):
    idle: bool
    nvidiaIndex: int
    nvidiaUUID: str
    allowToken: str


class ScheduleMetadata(BaseModel):
    auth: Optional[AuthLevel] = None
    queuing_reason: Optional[QueuingReason] = None


class QuotaInfos(BaseModel):
    left: int
    wait: timedelta


class QueueEvent(BaseModel):
    event: Literal['ping', 'failed', 'succeeded']
    data: Optional[ScheduleResponse] = None


def sse_parse(text: str):
    event, *data = text.strip().splitlines()
    assert event.startswith('event:')
    event = event[6:].strip()
    if event in ('ping', 'failed'):
        return QueueEvent(event=event)
    assert event == 'succeeded'
    (data,) = data
    assert data.startswith('data:')
    data = data[5:].strip()
    return QueueEvent(event=event, data=ScheduleResponse.parse_raw(data))


def sse_stream(res: httpx.Response) -> Generator[QueueEvent, Any, None]:
    for text in res.iter_text():
        if len(text) == 0:
            break # pragma: no cover
        try:
            yield sse_parse(text)
        except GeneratorExit:
            res.close()
            break


class APIClient:

    def __init__(self, client: httpx.Client):
        self.client = client

    def startup_report(self, cgroup_path: str, gpu_size: GPUSize) -> httpx.codes:
        res = self.client.post('/startup-report', params={
            'cgroupPath': cgroup_path,
            'gpuSize': gpu_size,
        })
        return httpx.codes(res.status_code)

    def schedule(
        self,
        cgroup_path: str,
        task_id: int = 0,
        token: str | None = None,
        token_version: int = 1,
        duration_seconds: int | None = None,
        enable_queue: bool = True,
    ):
        params: dict[str, str | int | bool] = {
            'cgroupPath': cgroup_path,
            'taskId': task_id,
            'enableQueue': enable_queue,
            'tokenVersion': token_version,
        }
        if duration_seconds is not None:
            params['durationSeconds'] = duration_seconds
        if token is not None:
            params['token'] = token
        res = self.client.send(
            request=self.client.build_request(
                method='POST',
                url='/schedule',
                params=params,
            ),
            stream=True,
        )
        status = httpx.codes(res.status_code)
        auth: AuthLevel | None = res.headers.get(AUTHENTICATED_HEADER)
        queuing_reason: QueuingReason | None = res.headers.get(QUEUING_REASON_HEADER)
        metadata = ScheduleMetadata(auth=auth, queuing_reason=queuing_reason)
        if (status is not httpx.codes.OK and
            status is not httpx.codes.TOO_MANY_REQUESTS
        ):
            res.close()
            return status, metadata
        if "text/event-stream" in res.headers['content-type']:
            return sse_stream(res), metadata
        res.read()
        if status is httpx.codes.TOO_MANY_REQUESTS:
            return QuotaInfos(**res.json()), metadata # pragma: no cover
        if status is httpx.codes.OK:
            return ScheduleResponse(**res.json()), metadata
        assert_never(status)

    def allow(
        self,
        allow_token: str,
        pid: int,
    ):
        res = self.client.post('/allow', params={
            'allowToken': allow_token,
            'pid': pid,
        })
        return httpx.codes(res.status_code)

    def release(
        self,
        allow_token: str,
        fail: bool = False,
    ) -> httpx.codes:
        res = self.client.post('/release', params={
            'allowToken': allow_token,
            'fail': fail,
        })
        return httpx.codes(res.status_code)

    def get_queue_size(self) -> float:
        res = self.client.get('/queue-size')
        assert res.status_code == 200, res.status_code
        size = res.json()
        return size



"""
Applied https://github.com/breuleux/jurigged/pull/58
"""

import __future__

import ast
import re
import sys
import traceback
from abc import abstractmethod
from ast import _splitlines_no_ff as _splitlines
from collections import Counter
from contextlib import contextmanager, suppress
from contextvars import ContextVar
from dataclasses import dataclass, field, replace as dc_replace
from types import CodeType, ModuleType
from typing import List, Optional, Union

from codefind import ConformException, code_registry as codereg, conform
from ovld import ovld, recurse

from .parse import Variables, variables
from .utils import EventSource, shift_lineno

current_info = ContextVar("current_info", default=None)
_future_feature_names = set(__future__.all_feature_names)


if sys.version_info < (3, 12):  # pragma: no cover
    splitlines = _splitlines
else:  # pragma: no cover

    def splitlines(s):
        lines = _splitlines(s)
        if len(lines) > 0 and lines[-1] == "":
            return lines[:-1]
        return lines


sep_at_start = re.compile(r"^ *[\n;]")
sep_at_end = re.compile(r"[\n;] *$")


class StaleException(Exception):
    pass


class attrproxy:
    def __init__(self, cls):
        self.cls = cls

    def __getitem__(self, item):
        try:
            return getattr(self.cls, item)
        except AttributeError:
            raise KeyError(item)

    def __setitem__(self, item, value):
        return setattr(self.cls, item, value)

    def get(self, item, dflt):
        return getattr(self.cls, item, dflt)


def _get_future_compiler_flags(glb) -> int:
    """use future feature names to get any that are set in the global namespace

    `|` these together to get the flags for the compile function

    see the source code for the `__future__` module for more info
    """
    flags = 0
    for key in glb.keys() & _future_feature_names:
        with suppress(Exception):
            flags |= glb[key].compiler_flag
    return flags


def _compile_with_flags(node, mode, filename, glb, flags=0):
    """compile the node with the future flags from the global namespace

    for now that means respect `from __future__ import annotations` if it exists
    """
    flags |= _get_future_compiler_flags(glb)
    return compile(node, mode=mode, filename=filename, flags=flags)


@dataclass
class Info:
    filename: str
    module_name: str
    source: str
    lines: list
    varinfo: Variables = None

    replace = dc_replace

    def get_segment(self, ext):
        lineno = ext.lineno - 1
        col_offset = ext.col_offset
        end_lineno = ext.end_lineno - 1
        end_col_offset = ext.end_col_offset

        lines = self.lines
        if end_lineno == lineno:
            return lines[lineno].encode()[col_offset:end_col_offset].decode()

        first = lines[lineno].encode()[col_offset:].decode()
        last = lines[end_lineno].encode()[:end_col_offset].decode()
        lines = lines[lineno + 1 : end_lineno]

        lines.insert(0, first)
        lines.append(last)
        return "".join(lines)


@contextmanager
def use_info(**fields):
    info = Info(**fields)
    token = current_info.set(info)
    try:
        yield
    finally:
        current_info.reset(token)


def get_info():
    return current_info.get()


@dataclass
class Correspondence:
    original: "Definition"
    new: "Definition"
    corresponds: bool
    changed: bool = False
    child_correspondences: Optional[List["Correspondence"]] = None

    @staticmethod
    def invalid(original, new):
        return Correspondence(
            original=original,
            new=new,
            corresponds=False,
            changed=False,
            child_correspondences=None,
        )

    @staticmethod
    def valid(original, new, **kwargs):
        return Correspondence(
            original=original,
            new=new,
            corresponds=True,
            **kwargs,
        )

    def fitness(self):
        return (
            int(self.corresponds),
            1 - int(self.changed),
        )

    def walk(self):
        yield self
        for child in self.child_correspondences or []:
            yield from child.walk()

    def summary(self, filter=None):
        (same, changes, additions, deletions) = ([], [], [], [])
        for corr in self.walk():
            if filter is None or filter(corr.original or corr.new):
                if corr.original is None:
                    additions.append(corr.new)
                elif corr.new is None:
                    deletions.append(corr.original)
                elif corr.changed:
                    changes.append(corr.original)
                else:
                    same.append(corr.original)
        return (same, changes, additions, deletions)


@dataclass
class Definition:
    node: ast.AST
    name: str = None
    filename: str = None
    parent: Optional["Definition"] = None

    # This is the original line number, used in the first lookup of the
    # code object. It does not need to remain in sync with updates to the
    # source code.
    groundline: int = -1

    def __post_init__(self):
        self._code = None
        if self.filename is None:
            self.filename = get_info().filename

    #############
    # Hierarchy #
    #############

    def set_parent(self, parent):
        self.parent = parent
        for p in self.hierarchy(skip=1):
            p._code = None

    def hierarchy(self, skip=0):
        if skip <= 0:
            yield self
        if self.parent is not None:
            yield from self.parent.hierarchy(skip - 1)

    def dotpath(self):
        chain = list(self.hierarchy())
        return ".".join(x.name or "<line>" for x in reversed(chain))

    def codepath(self, skip=0):
        chain = list(self.hierarchy(skip=skip))
        return tuple(
            (x.filename if i == 0 else x.name) or "<line>"
            for i, x in enumerate(reversed(chain))
        )

    def get_globals(self):
        return self.parent and self.parent.get_globals()

    def get_object(self):
        return None

    def walk(self):
        yield self

    ##############
    # Management #
    ##############

    @property
    def codestring(self):
        if self._code is None:
            self._code = self.reconstruct()
        return self._code

    @property
    def is_whitespace(self):
        return False

    @abstractmethod
    def reconstruct(self):
        pass

    @abstractmethod
    def stash(self, lineno=1, col_offset=0):
        pass

    @abstractmethod
    def prepend_text(self, text):
        pass

    @abstractmethod
    def append_text(self, text):
        pass

    ##################
    # Correspondence #
    ##################

    @abstractmethod
    def correspond(self, other):
        pass

    @abstractmethod
    def apply_correspondence(self, corr, order, controller):
        pass

    ##############
    # Evaluation #
    ##############

    def evaluate(self, glb, lcl):
        if self.node is not None:
            node = ast.Module(body=[self.node], type_ignores=[])
            code = _compile_with_flags(
                node, mode="exec", filename=self.filename, glb=glb
            )
            code = code.replace(co_name="<adjust>")
            exec(code, glb, lcl)
            codereg.assimilate(
                code.replace(co_name=""), path=self.codepath(skip=1)
            )

    #############
    # Utilities #
    #############

    def well_separated(self, other):
        a = self.codestring
        b = other.codestring
        return sep_at_end.search(a) or sep_at_start.search(b)


@dataclass
class LineDefinition(Definition):
    text: str = ""

    ##############
    # Management #
    ##############

    def reconstruct(self):
        return self.text

    def stash(self, lineno=1, col_offset=0):
        lines = self.text.split("\n")
        last = len(lines[-1])
        self.stashed = Extent(
            lineno=lineno,
            col_offset=col_offset,
            end_lineno=lineno + len(lines) - 1,
            end_col_offset=col_offset + last if len(lines) == 1 else last,
            filename=self.filename,
            content=self.codestring,
        )
        return self.stashed

    def prepend_text(self, text):
        self.text = text + self.text

    def append_text(self, text):
        self.text = self.text + text

    @property
    def is_whitespace(self):
        return not any(substantial(line) for line in self.text.split("\n"))

    ##################
    # Correspondence #
    ##################

    def equiv_src(self, other):
        return self.text == other.text

    def correspond(self, other):
        if type(other) is not type(self) or not self.equiv_src(other):
            return Correspondence.invalid(self, other)
        else:
            return Correspondence.valid(self, other, changed=False)


@dataclass
class HeaderDefinition(LineDefinition):
    ##################
    # Correspondence #
    ##################

    def equiv_src(self, other):
        return self.text.strip() == other.text.strip()


@dataclass
class GroupDefinition(Definition):
    variables: Variables = None
    children: List[Definition] = field(default=list)

    def __post_init__(self):
        super().__post_init__()
        self.ignore_names = False
        children, self.children = self.children, []
        for child in children:
            self.append(child)

    #############
    # Hierarchy #
    #############

    def set_parent(self, parent):
        super().set_parent(parent)
        if self.variables is not None:
            closable = set()
            for p in self.hierarchy(skip=1):
                if p.variables:
                    closable |= p.variables.assigned
            self.variables.closure = self.variables.free & closable

    def header(self):
        return "".join(
            [
                child.codestring
                for child in self.children
                if isinstance(child, HeaderDefinition)
            ]
        )

    def walk(self):
        yield self
        for child in self.children:
            yield from child.walk()

    ##############
    # Management #
    ##############

    def reconstruct(self):
        return "".join([child.codestring for child in self.children])

    def stash(self, lineno=1, col_offset=0):
        self.stashed = Extent(
            lineno=lineno,
            col_offset=col_offset,
            end_lineno=lineno,
            end_col_offset=col_offset,
            filename=self.filename,
            content=self.codestring,
        )
        curr = self.stashed
        for child in self.children:
            curr = child.stash(curr.end_lineno, curr.end_col_offset)
        self.stashed.end_lineno = curr.end_lineno
        self.stashed.end_col_offset = curr.end_col_offset
        return self.stashed

    def prepend_text(self, text):
        if self.children:
            self.children[0].prepend_text(text)
        else:  # pragma: no cover
            # This doesn't seem to ever happen
            self.prepend(
                LineDefinition(node=None, text=text, filename=self.filename)
            )

    def append_text(self, text):  # pragma: no cover
        # This doesn't seem to ever be called
        if self.children:
            self.children[-1].append_text(text)
        else:
            self.append(
                LineDefinition(node=None, text=text, filename=self.filename)
            )

    def append(self, *children, ensure_separation=False):
        for child in children:
            if (
                ensure_separation
                and self.children
                and not self.children[-1].well_separated(child)
            ):
                ws = LineDefinition(
                    node=None, text="\n", filename=self.filename
                )
                self.children.append(ws)
                ws.set_parent(self)
            self.children.append(child)
            child.set_parent(self)

    def prepend(self, *children):
        self.children[0:0] = children
        for child in children:
            child.set_parent(self)

    ##################
    # Correspondence #
    ##################

    def correspond(self, other):
        if type(other) is not type(self) or (
            not self.ignore_names and self.name != other.name
        ):
            return Correspondence.invalid(self, other)
        elif self.codestring == other.codestring:
            return Correspondence.valid(self, other, changed=False)
        else:
            childcorr = []
            children = list(self.children)

            for other_child in other.children:
                candidates = [
                    corr
                    for this_child in children
                    if (corr := this_child.correspond(other_child)).corresponds
                ]

                if not candidates:
                    corr = Correspondence.valid(None, other_child, changed=True)
                else:
                    corr = max(candidates, key=lambda corr: corr.fitness())
                    children.remove(corr.original)

                childcorr.append(corr)

            for child in children:
                corr = Correspondence.valid(child, None, changed=True)
                childcorr.append(corr)

            mergeable = not any(
                (
                    isinstance(corr.original, HeaderDefinition)
                    or isinstance(corr.new, HeaderDefinition)
                )
                and corr.changed
                for corr in childcorr
            )

            if mergeable:
                return Correspondence.valid(
                    original=self,
                    new=other,
                    changed=True,
                    child_correspondences=childcorr,
                )
            else:
                return Correspondence.invalid(self, other)

    def _process_child_correspondence(self, ccorr, order, controller):
        orig = ccorr.original
        new = ccorr.new

        try:
            if orig is None:
                if controller("pre-add", ccorr):
                    # Addition
                    self.append(new, ensure_separation=True)
                    self.evaluate_child(new)
                    controller("post-add", ccorr)
            elif new is None:
                if controller("pre-delete", ccorr):
                    # Deletion
                    conform(orig.get_object(), None)
                    controller("post-delete", ccorr)
                else:
                    self.append(orig, ensure_separation=True)
            elif ccorr.changed:
                # Change
                self.append(orig, ensure_separation=True)
                try:
                    orig.apply_correspondence(
                        ccorr,
                        order=order,
                        controller=controller,
                    )
                except ConformException:
                    self.children.pop()
                    self._process_child_correspondence(
                        Correspondence.valid(None, new),
                        order=order,
                        controller=controller,
                    )
            else:
                self.append(orig, ensure_separation=True)
        except Exception as exc:
            controller("error", ccorr, exc=exc)

    def _apply_corrlist(self, corrs, order, controller):
        def namecounts():
            c = Counter()
            for child in self.children:
                if (name := getattr(child, "name", None)) is not None:
                    c[name] += 1
            return c

        counts1 = namecounts()
        self.children = []

        for corr in corrs:
            self._process_child_correspondence(corr, order, controller)

        counts2 = namecounts()
        for dlt in set(counts1) - set(counts2):
            self.delete_property(dlt)

    def _apply_correspondence_orig_order(self, corr, controller):
        groups = {id(None): []}
        curr = None
        for ccorr in corr.child_correspondences:
            if ccorr.original is not None:
                if ccorr.original.node is not None:
                    if curr is None:
                        init = groups[id(None)]
                        del groups[id(None)]
                    else:
                        init = []
                    curr = ccorr.original
                    groups[id(curr)] = init
                    groups[id(curr)].append(ccorr)
                else:
                    groups[id(ccorr.original)] = [ccorr]
            else:
                groups[id(curr)].append(ccorr)

        ccorrs = []
        for child in self.children:
            ccorrs += groups.get(id(child), [])
        ccorrs += groups.get(id(None), [])
        self._apply_corrlist(
            ccorrs,
            order="original",
            controller=controller,
        )

    def _apply_correspondence_new_order(self, corr, controller):
        self._apply_corrlist(
            corr.child_correspondences, order="new", controller=controller
        )

    def apply_correspondence(self, corr, order, controller):
        assert corr.corresponds

        if not corr.changed:
            return

        if controller("pre-update", corr):
            assert order in ("original", "new")

            if order == "original":
                self._apply_correspondence_orig_order(
                    corr, controller=controller
                )
            elif order == "new":
                self._apply_correspondence_new_order(
                    corr, controller=controller
                )

            controller("post-update", corr)

    ##############
    # Evaluation #
    ##############

    def evaluate(self, glb, lcl):
        super().evaluate(glb, lcl)
        obj = (lcl or glb).get(self.name, None)
        if hasattr(obj, "__qualname__"):
            obj.__qualname__ = ".".join(self.dotpath().split(".")[1:])

    @abstractmethod
    def evaluate_child(self, child):
        pass

    @abstractmethod
    def delete_property(self, prop):
        pass


@dataclass
class ModuleCode(GroupDefinition):
    module: object = None
    globals: object = None

    def __post_init__(self):
        super().__post_init__()
        self.ignore_names = True

    #############
    # Hierarchy #
    #############

    def get_globals(self):
        return self.globals

    def get_object(self):
        return self.globals

    ##############
    # Evaluation #
    ##############

    def evaluate_child(self, child):
        return child.evaluate(self.get_globals(), None)

    def delete_property(self, prop):
        del self.globals[prop]


@dataclass
class ClassDefinition(GroupDefinition):
    ##############
    # Evaluation #
    ##############

    def get_object(self):
        parent = self.parent.get_object()
        if isinstance(parent, dict):
            return parent.get(self.name, None)
        else:
            return getattr(parent, self.name, None)

    def evaluate_child(self, child):
        if (obj := self.get_object()) is not None:
            return child.evaluate(self.get_globals(), attrproxy(obj))

    def delete_property(self, prop):
        if (obj := self.get_object()) is not None:
            delattr(obj, prop)


@dataclass
class FunctionDefinition(GroupDefinition):
    _codeobj: object = None

    ##############
    # Management #
    ##############

    def stash(self, lineno=1, col_offset=0):
        if not isinstance(self.parent, FunctionDefinition):
            co = self.get_object()
            if co and (delta := lineno - co.co_firstlineno):
                self.recode(shift_lineno(co, delta), use_cache=False)

        return super().stash(lineno, col_offset)

    ##################
    # Correspondence #
    ##################

    def recode(self, new_code, recode_current=True, use_cache=False):
        # Gather the code objects of all closures into subcodes
        subcodes = {}

        def _fill_subcodes(code, path):
            subcodes[path] = code
            for co in code.co_consts:
                if isinstance(co, CodeType):
                    _fill_subcodes(co, (*path, co.co_name))

        here = self.codepath()
        _fill_subcodes(new_code, here)
        if not recode_current:
            del subcodes[here]

        # Synchronize changes in closure codes
        for closure in self.walk():
            if isinstance(closure, FunctionDefinition) and (
                subcode := subcodes.get(closure.codepath(), None)
            ):
                co = closure.get_object()
                if co is not subcode:
                    conform(co, subcode, use_cache=use_cache)
                    closure._codeobj = subcode

    def apply_correspondence(self, corr, order, controller):
        assert corr.corresponds and corr.changed

        if controller("pre-update", corr):
            # Reevaluate this function
            glb = self.get_globals()
            new_obj = self.reevaluate(corr.new.node, glb)
            new_code = new_obj.__code__

            self.recode(new_code, recode_current=False)

            # We will throw out all original child correspondences and replace
            # them by the new, so if the reevaluation succeeds it is important
            # to sync their code objects.
            for ccorr in corr.walk():
                if (
                    isinstance(ccorr.original, FunctionDefinition)
                    and ccorr.new is not None
                ):
                    ccorr.new._codeobj = ccorr.original._codeobj

            self.children = []
            self.append(*corr.new.children)
            self._codeobj = new_code
            controller("post-update", corr)

    ##############
    # Evaluation #
    ##############

    def get_object(self):
        if self._codeobj is None:
            pth = (*self.codepath(), self.groundline)
            if pth in codereg.codes:
                self._codeobj = codereg.codes[pth]
        return self._codeobj

    def reevaluate(self, new_node, glb):
        ext = new_node.extent
        closure = False
        lcl = {}
        new_node = type(new_node)(
            name=new_node.name,
            args=new_node.args,
            body=new_node.body,
            decorator_list=[],
            returns=new_node.returns,
            type_comment=new_node.type_comment,
            lineno=new_node.lineno,
            col_offset=new_node.col_offset,
            end_lineno=new_node.end_lineno,
            end_col_offset=new_node.end_col_offset,
        )
        previous = lcl.get(self.name, None)
        if self.variables.closure:
            # Because reevaluate is typically not run on closures, this code
            # path is essentially only entered for functions that use super(),
            # since they are implicit closures on __class__
            closure = True
            names = tuple(sorted(self.variables.closure))
            wrap = ast.copy_location(
                ast.FunctionDef(
                    name="##create_closure",
                    args=ast.arguments(
                        posonlyargs=[],
                        args=[
                            ast.arg(
                                arg=name, lineno=new_node.lineno, col_offset=0
                            )
                            for name in names
                        ],
                        vararg=None,
                        kwonlyargs=[],
                        kw_defaults=[],
                        kwarg=None,
                        defaults=[],
                    ),
                    body=[
                        new_node,
                        ast.Return(ast.Name(id=new_node.name, ctx=ast.Load())),
                    ],
                    decorator_list=[],
                    returns=None,
                ),
                new_node,
            )
            ast.fix_missing_locations(wrap)
            node = ast.Module(body=[wrap], type_ignores=[])
        else:
            node = ast.Module(body=[new_node], type_ignores=[])
        code = _compile_with_flags(
            node, mode="exec", filename=ext.filename, glb=glb
        )
        code = code.replace(co_name="<adjust>")
        exec(code, glb, lcl)
        if closure:
            creator = lcl["##create_closure"]
            # It does not matter what arguments we provide here, because we will move the
            # function's __code__ elsewhere, so it will use a different closure
            new_obj = creator(*names)
        else:
            new_obj = lcl[self.name]
        lcl[self.name] = previous
        node.extent = ext
        self.node = node
        old_obj = self.get_object()
        if old_obj is None:  # pragma: no cover
            raise Exception(
                f"Cannot find existing object for replacement of '{self.name}'."
            )
        conform(old_obj, new_obj)
        self._codeobj = new_obj.__code__
        return new_obj


@dataclass
class Extent:
    lineno: int
    col_offset: int
    end_lineno: int
    end_col_offset: int
    filename: str = None
    content: str = None

    def __post_init__(self):
        if self.filename is None:
            self.filename = get_info().filename


def _collapse_to_beginning(ext):
    return Extent(
        lineno=ext.lineno,
        col_offset=ext.col_offset,
        end_lineno=ext.lineno,
        end_col_offset=ext.col_offset,
    )


def _collapse_to_end(ext):
    return Extent(
        lineno=ext.end_lineno,
        col_offset=ext.end_col_offset,
        end_lineno=ext.end_lineno,
        end_col_offset=ext.end_col_offset,
    )


def extend_to_line(node):
    return Extent(
        lineno=node.lineno,
        col_offset=0,
        end_lineno=node.end_lineno,
        end_col_offset=node.end_col_offset,
    )


def fill_real_extent(node):
    extents = [
        ext for n in ast.iter_child_nodes(node) if (ext := fill_real_extent(n))
    ]
    if hasattr(node, "decorator_list"):
        for deco in node.decorator_list:
            extents.append(extend_to_line(deco))

    if hasattr(node, "lineno"):
        extents.append(node)

        lineno, col_offset = min(
            (ext.lineno, ext.col_offset) for ext in extents
        )
        end_lineno, end_col_offset = max(
            (ext.end_lineno, ext.end_col_offset) for ext in extents
        )
        node.extent = Extent(
            lineno=lineno,
            col_offset=col_offset,
            end_lineno=end_lineno,
            end_col_offset=end_col_offset,
        )
    else:
        node.extent = None
    return node.extent


def substantial(s):
    return not re.fullmatch(r" *(#.*)?\n?", s)


def analyze_split(s):
    lines = splitlines(s)
    subst = max(
        [i for i, line in enumerate(lines) if substantial(line)], default=-1
    )
    left = lines[: subst + 1]
    middle = lines[subst + 1 :]
    if middle and not middle[-1].endswith("\n"):
        right = [middle.pop()]
    else:
        right = []
    return "".join(left), "".join(middle), "".join(right)


def delta(node1, node2):
    return get_info().get_segment(
        Extent(
            lineno=node1.end_lineno,
            col_offset=node1.end_col_offset,
            end_lineno=node2.lineno,
            end_col_offset=node2.col_offset,
        )
    )


def distribute(between, defn1, defn2, cls=LineDefinition):
    left, middle, right = analyze_split(between)
    rval = ""
    if left:
        if defn1:
            defn1.append_text(left)
        else:
            rval += left
    if middle:
        rval += middle
    if right:
        if defn2:
            defn2.prepend_text(right)
        else:
            rval += right
    return [cls(node=None, text=rval)] if rval else []


@ovld
def collect_definitions(nodes: list):
    if not nodes:
        return []
    defns = [(node.extent, recurse(node)) for node in nodes]
    results = []
    for (node1, defn1), (node2, defn2) in zip(defns[:-1], defns[1:]):
        between = delta(node1, node2)
        results.append(defn1)
        results.extend(distribute(between, defn1, defn2))
    results.append(defns[-1][1])
    return results


@ovld
def collect_definitions(node: Union[ast.FunctionDef, ast.AsyncFunctionDef]):
    info = get_info()
    defns = recurse(node.body)
    fndefn = FunctionDefinition(
        name=node.name,
        node=node,
        children=defns,
        variables=info.varinfo.get(node, Variables()).replace(),
    )

    prelude = []

    deco0 = _collapse_to_beginning(node.extent)
    between = delta(deco0, node)
    prelude += distribute(between, None, None, cls=HeaderDefinition)

    fnstart = _collapse_to_beginning(node)
    between = delta(fnstart, node.body[0].extent)
    prelude += distribute(between, None, defns[0])

    fndefn.prepend(*prelude)

    fnend = _collapse_to_end(node)
    between = delta(node.body[-1].extent, fnend)
    fndefn.append(*distribute(between, defns[-1], None))

    fndefn.groundline = deco0.lineno

    return fndefn


@ovld
def collect_definitions(node: ast.ClassDef):
    info = get_info()
    defns = recurse(node.body)
    clsdefn = ClassDefinition(
        name=node.name,
        node=node,
        children=defns,
        variables=info.varinfo.get(node, Variables()).replace(),
    )

    prelude = []

    deco0 = _collapse_to_beginning(node.extent)
    between = delta(deco0, node.body[0].extent)
    prelude += distribute(between, None, defns[0], cls=HeaderDefinition)

    clsdefn.prepend(*prelude)

    fnend = _collapse_to_end(node)
    between = delta(node.body[-1].extent, fnend)
    clsdefn.append(*distribute(between, defns[-1], None))

    return clsdefn


@ovld
def collect_definitions(node: ast.Module):
    info = get_info()
    begin_node = Extent(lineno=1, col_offset=0, end_lineno=1, end_col_offset=0)
    end_node = Extent(
        lineno=len(info.lines),
        col_offset=len(info.lines[-1]),
        end_lineno=len(info.lines),
        end_col_offset=len(info.lines[-1]),
    )

    cg = ModuleCode(
        node=node, name=info.module_name, children=recurse(node.body)
    )

    if node.body:
        if between := delta(begin_node, node.body[0].extent):
            cg.prepend(*distribute(between, None, None))

        if between := delta(node.body[-1].extent, end_node):
            cg.append(*distribute(between, None, None))

    return cg


@ovld
def collect_definitions(node: ast.stmt):
    return LineDefinition(node=node, text=get_info().get_segment(node))


class CodeFile:
    def __init__(self, filename, module_name, source=None):
        self.activity = EventSource()
        self.filename = filename
        # if not self.filename.startswith("/") and not self.filename.startswith("<"):
        #     self.filename = os.path.abspath(self.filename)
        self.module_name = module_name
        self.saved = self.read_source() if source is None else source
        if not self.saved.endswith("\n"):
            self.saved += "\n"
        tree = ast.parse(self.saved)
        varinfo = {}
        variables(tree, varinfo)
        with use_info(
            filename=self.filename,
            module_name=module_name,
            source=self.saved,
            lines=splitlines(self.saved),
            varinfo=varinfo,
        ):
            fill_real_extent(tree)
            self.root = collect_definitions(tree)
        self.root.stash()
        self.dirty = False

    @property
    def module(self):
        return self.root.module

    def associate(self, obj):
        if isinstance(obj, ModuleType):
            self.root.module = obj
            self.root.globals = vars(obj)
        elif isinstance(obj, dict):
            self.root.module = None
            self.root.globals = obj
        else:
            raise TypeError("associate expects a dict or module")

    def read_source(self):
        with open(self.filename) as f:
            source = f.read()
        if not source.endswith("\n"):
            source += "\n"
        return source

    def stale(self):
        return self.read_source() != self.saved

    def merge(self, other, order="original", allow_deletions=True):
        if order == "new":
            assert allow_deletions

        def controller(op, ccorr, exc=None):
            if op == "pre-delete":
                return allow_deletions and (
                    allow_deletions is True or ccorr.original in allow_deletions
                )
            elif op == "post-add":
                if not ccorr.new.is_whitespace:
                    self.activity.emit(AddOperation(self, ccorr.new))
            elif op == "post-delete":
                if not ccorr.original.is_whitespace:
                    self.activity.emit(DeleteOperation(self, ccorr.original))
            elif op == "post-update":
                self.activity.emit(UpdateOperation(self, ccorr.original))
            elif op == "error":
                self.activity.emit(ExceptionOperation(self, ccorr.new, exc))
            else:
                return True

        corr = self.root.correspond(other.root)
        if corr.changed:
            self.dirty = True
        self.root.apply_correspondence(corr, order=order, controller=controller)
        return corr.summary()

    def commit(self, check_stale=True):
        if not self.dirty:
            return
        if check_stale and self.stale():
            raise StaleException(
                f"Cannot commit changes to {self.filename} because the file was changed."
            )
        new_source = self.root.reconstruct()
        if not new_source.endswith("\n"):
            new_source += "\n"
        with open(self.filename, "w") as f:
            f.write(new_source)
        self.root.stash()
        self.saved = new_source
        self.dirty = False

    def refresh(self):
        new_source = self.read_source()
        if new_source != self.root.codestring or self.dirty:
            cf = CodeFile(
                self.filename, source=new_source, module_name=self.module_name
            )
            self.merge(cf, order="new")
            self.root.stash()


@dataclass
class CodeFileOperation:
    codefile: CodeFile
    defn: Definition


@dataclass
class UpdateOperation(CodeFileOperation):
    def __str__(self):
        return f"Update {self.defn.dotpath()} @L{self.defn.stashed.lineno}"


@dataclass
class AddOperation(CodeFileOperation):
    def __str__(self):
        if isinstance(self.defn, LineDefinition):
            return f"Run {self.defn.parent.dotpath()} @L{self.defn.stashed.lineno}: {self.defn.text}"
        else:
            return f"Add {self.defn.dotpath()} @L{self.defn.stashed.lineno}"


@dataclass
class ExceptionOperation(CodeFileOperation):
    exc: Optional[Exception]
    def __str__(self):
        if isinstance(self.defn, LineDefinition):
            msg = f"Exception (run) {self.defn.parent.dotpath()} @L{self.defn.stashed.lineno}: {self.defn.text}"
        else:
            msg = f"Exception (add) {self.defn.dotpath()} @L{self.defn.stashed.lineno}"
        if self.exc is None:
            return msg
        tb = "".join(traceback.format_exception(type(self.exc), self.exc, self.exc.__traceback__))
        return f"{msg}\n{tb}"


@dataclass
class DeleteOperation(CodeFileOperation):
    def __str__(self):
        return f"Delete {self.defn.dotpath()} @L{self.defn.stashed.lineno}"



"""
"""
from __future__ import annotations

import os
import time
import warnings
from datetime import timedelta
from typing import Any

import gradio as gr
import httpx
from packaging import version
from typing_extensions import assert_never

from ..config import Config
from . import utils
from .api import APIClient
from .api import GPUSize
from .api import AuthLevel
from .api import QuotaInfos
from .api import ScheduleResponse
from .gradio import info
from .gradio import error
from .gradio import get_event
from .gradio import supports_auth


TOKEN_HEADER = 'X-IP-Token'
DEFAULT_SCHEDULE_DURATION = 60

UNUSED_MESSAGE = "GPU device not used"
NO_GPU_MESSAGE_REGULAR = "No GPU was available"
NO_GPU_MESSAGE_INQUEUE = "No GPU was available after 60s"
EXAMPLES_RETRY_MESSAGE = "Try re-running outside of examples if it happened after clicking one"

SIGNUP_ON_HF_TXT = "Create a free account"
SIGNUP_ON_HF_URL = "https://huggingface.co/join"
SUBSCRIBE_TO_PRO_TXT = "Subscribe to Pro"
SUBSCRIBE_TO_PRO_URL = "https://huggingface.co/settings/billing/subscription"


def api_client():
    assert Config.zero_device_api_url is not None
    httpx_client = httpx.Client(base_url=Config.zero_device_api_url, timeout=60, verify=False)
    return APIClient(httpx_client)


def startup_report(cgroup_path: str, gpu_size: GPUSize):
    retries, max_retries = 0, 2
    client = api_client()
    while (status := client.startup_report(cgroup_path, gpu_size)) is httpx.codes.NOT_FOUND: # pragma: no cover
        time.sleep(1)
        if (retries := retries + 1) > max_retries:
            raise RuntimeError("Error while initializing ZeroGPU: NotFound")
    if status is httpx.codes.CONFLICT: # pragma: no cover (quickfix: should be fixed at device-api level)
        return
    if status is not httpx.codes.OK: # pragma: no cover
        raise RuntimeError("Error while initializing ZeroGPU: Unknown")


def html_string(html_contents: str, text_contents: str): # pragma: no cover
    class HTMLString(str):
        def __str__(self):
            return text_contents
    return HTMLString(html_contents)


def _toast_action(
    auth: AuthLevel | None,
    supports_html: bool,
    pro_message: str,
    unlogged_desc: str,
    logged_desc: str,
    ending: str,
) -> tuple[str, str]: # pragma: no cover
    if not supports_auth() or auth == 'pro':
        return pro_message, pro_message
    html = ""
    link = SIGNUP_ON_HF_URL if auth is None else SUBSCRIBE_TO_PRO_URL
    text = SIGNUP_ON_HF_TXT if auth is None else SUBSCRIBE_TO_PRO_TXT
    desc = unlogged_desc if auth is None else logged_desc
    desc += f" {ending}."
    style = ";".join([
        "white-space: nowrap",
        "text-underline-offset: 2px",
        "color: var(--body-text-color)",
    ])
    if supports_html:
        html += f'<a style="{style}" href="{link}">'
    html += text
    if supports_html:
        html += '</a>'
    html += f" {desc}"
    markdown = f'[{text}]({link}) {desc}'
    return html, markdown


def schedule(
    task_id: int,
    request: gr.Request | None = None,
    duration: timedelta | None = None,
    _first_attempt: bool = True,
) -> ScheduleResponse:

    if not (gradio_version := version.parse(gr.__version__)).major >= 4: # pragma: no cover
        raise RuntimeError("ZeroGPU is only compatible with Gradio 4+")

    GRADIO_HTML_TOASTS = gradio_version >= version.Version('4.39')
    GRADIO_HANDSHAKE = gradio_version >= version.Version('5.16.1')

    token, payload = _get_token_and_payload(request)
    if token is not None and (token_error := payload.get('error')):
        message = f"Falling back to IP-based quotas ({token_error})"
        info("ZeroGPU client warning", message, level='warning')

    res, meta = api_client().schedule(
        cgroup_path=utils.self_cgroup_device_path(),
        task_id=task_id,
        token=token,
        token_version=2 if GRADIO_HANDSHAKE else 1,
        duration_seconds=duration.seconds if duration is not None else None,
    )

    auth = meta.auth

    if isinstance(res, ScheduleResponse):
        return res

    if isinstance(res, QuotaInfos): # pragma: no cover
        requested = duration.seconds if duration is not None else DEFAULT_SCHEDULE_DURATION
        if res.wait < timedelta(0):
            message = (
                f"The requested GPU duration ({requested}s) "
                f"is larger than the maximum allowed"
            )
            raise error("ZeroGPU illegal duration", message)
        elif token is None:
            message = (
                f"Space app has reached its GPU limit. "
                f"{EXAMPLES_RETRY_MESSAGE}"
            )
            raise error("ZeroGPU quota exceeded", message)
        else:
            if payload.get('user') is None and res.wait == 0:
                message = "You have exceeded your runs limit."
            else:
                gpu = "Pro GPU" if auth == 'pro' else ("free GPU" if auth == 'regular' else "GPU")
                message_gui = (
                    f"You have exceeded your {gpu} quota "
                    f"({requested}s requested vs. {res.left}s left). "
                    f"Try again in {res.wait}"
                )
                if auth is None:
                    message_mcp = (
                        "Unlogged user is runnning out of daily ZeroGPU quotas. "
                        "Signup for free on https://huggingface.co/join "
                        "or login on https://huggingface.co/login "
                        "to get more ZeroGPU quota now."
                    )
                elif auth == 'regular':
                    message_mcp = (
                        "User is runnning out of daily ZeroGPU quotas. "
                        "Visit https://huggingface.co/subscribe/pro "
                        "to get more ZeroGPU quota now."
                    )
                else:
                    message_mcp = message_gui
                message = html_string(message_gui, message_mcp)
            raise error("ZeroGPU quota exceeded", message, html=True)

    if not isinstance(res, httpx.codes): # pragma: no cover
        if meta.queuing_reason in ('node', None):
            info("ZeroGPU queue", "Waiting for a GPU to become available")
        elif meta.queuing_reason == 'concurrency':
            info("ZeroGPU queue", "Waiting for a GPU slot on this Space")
        else:
            assert_never(meta.queuing_reason)
        # TODO: Sign-up message if not authenticated (after some time ?)
        connection_event = get_event()
        if connection_event is None and request is not None:
            warnings.warn("ZeroGPU: Cannot get Gradio app Queue instance")
        while True:
            try:
                event = next(res)
            except StopIteration:
                raise RuntimeError("Unexpected end of stream")
            except httpx.RemoteProtocolError:
                if not _first_attempt:
                    raise RuntimeError("Error while re-trying after queue disconnect")
                return schedule(task_id, request, duration, _first_attempt=False)
            if event.event == 'ping':
                if connection_event is not None and not connection_event.alive:
                    res.close()
                    raise RuntimeError("Connection closed by visitor while queueing")
                continue
            if event.event == 'failed':
                if token is None:
                    message = f"{NO_GPU_MESSAGE_INQUEUE}. {EXAMPLES_RETRY_MESSAGE}"
                    raise error("ZeroGPU quota exceeded", message)
                details_html, details_markdown = _toast_action(
                    auth=auth,
                    supports_html=GRADIO_HTML_TOASTS,
                    pro_message="Retry later",
                    unlogged_desc="to get a higher",
                    logged_desc="to get the highest",
                    ending="priority in ZeroGPU queues",
                )
                message_html = f"{NO_GPU_MESSAGE_INQUEUE}. {details_html}"
                message_text = f"{NO_GPU_MESSAGE_INQUEUE} {details_markdown}"
                message = html_string(message_html, message_text)
                raise error("ZeroGPU queue timeout", message, html=True)
            if event.event == 'succeeded':
                assert event.data is not None
                if connection_event is not None and not connection_event.alive:
                    release(event.data.allowToken)
                    raise RuntimeError("Connection closed by visitor on queue success")
                info("ZeroGPU queue", "Successfully acquired a GPU", level='success')
                return event.data

    if res is httpx.codes.SERVICE_UNAVAILABLE:
        raise error("ZeroGPU client error", NO_GPU_MESSAGE_REGULAR)

    if res is httpx.codes.UNAUTHORIZED: # pragma: no cover
        raise error("ZeroGPU client error", "Expired ZeroGPU proxy token")

    # TODO: Find a way to log 'detail' response field
    raise RuntimeError(f"ZeroGPU API /schedule error: {res} ({httpx.codes.get_reason_phrase(res)})") # pragma: no cover


def allow(allow_token: str) -> None:
    pid = os.getpid()
    assert pid != 1, "Allowing PID 1 on ZeroGPU will end up killing your Space"
    assert api_client().allow(allow_token=allow_token, pid=pid) is httpx.codes.OK


def release(
    allow_token: str, *,
    fail: bool = False,
    allow_404: bool = False,
) -> None:

    res = api_client().release(
        allow_token=allow_token,
        fail=fail,
    )

    if res is httpx.codes.NO_CONTENT: # pragma: no cover
        try:
            info("ZeroGPU client warning", UNUSED_MESSAGE, level='warning')
        except AttributeError:
            pass
        warnings.warn(UNUSED_MESSAGE, RuntimeWarning)
        return None

    if res is httpx.codes.NOT_FOUND:
        if not allow_404:
            warnings.warn("ZeroGPU API /release warning: 404 Not Found")
        return None

    if httpx.codes.is_success(res):
        return None

    # TODO: Find a way to log 'detail' response field
    # TODO: Only raise in dev environment. Simply warn in production ?
    raise RuntimeError(f"ZeroGPU API /release error: {res} ({httpx.codes.get_reason_phrase(res)})") # pragma: no cover


def _get_token(request: gr.Request | None) -> str | None:

    if request is None:
        return None

    headers = getattr(request, 'headers', None)
    if headers is None or not hasattr(headers, '__dict__'):
        raise error("ZeroGPU client error", "Internal Gradio error")

    # Compatibility trick
    if not hasattr(headers, 'get'):
        headers = headers.__dict__ # pragma: no cover

    return headers.get(TOKEN_HEADER.lower())


def _get_token_and_payload(request: gr.Request | None) -> tuple[str | None, dict[str, Any]]:
    if (token := _get_token(request)) is None:
        return None, {}
    try:
        payload = utils.jwt_payload(token)
    except Exception: # pragma: no cover
        warnings.warn("Error while decoding X-IP-Token JWT")
        return token, {}
    return token, payload



"""
"""
# pyright: reportPrivateImportUsage=false

from __future__ import annotations

import multiprocessing
import os
from concurrent.futures import ProcessPoolExecutor
from contextlib import suppress
from functools import partial
from types import SimpleNamespace
from typing import Any
from typing import Callable
from typing import Optional
from typing import Tuple

import torch
from torch.utils.weak import WeakTensorKeyDictionary

from ...config import Config
from . import bitsandbytes


# Nvidia A100.80G MIG (drivers 535) / Torch 2.2.0
CUDA_DEVICE_NAME = 'NVIDIA A100-SXM4-80GB MIG 3g.40gb'
CUDA_TOTAL_MEMORY = 42144366592
CUDA_MEM_GET_INFO = (41911451648, CUDA_TOTAL_MEMORY)
CUDA_DEVICE_CAPABILITY = (8, 0)
CUDA_DEVICE_PROPERTIES = SimpleNamespace(name=CUDA_DEVICE_NAME, major=8, minor=0, total_memory=CUDA_TOTAL_MEMORY, multi_processor_count=42)

GENERIC_METHOD_NAMES = [
    'arange',
    'as_tensor',
    'asarray',
    'bartlett_window',
    'blackman_window',
    'empty',
    'empty_like',
    'empty_strided',
    'eye',
    'full',
    'full_like',
    'hamming_window',
    'hann_window',
    'kaiser_window',
    'linspace',
    'logspace',
    'ones',
    'ones_like',
    'rand',
    'rand_like',
    'randint',
    'randint_like',
    'randn',
    'randn_like',
    'randperm',
    'range',
    'sparse_bsc_tensor',
    'sparse_bsr_tensor',
    'sparse_compressed_tensor',
    'sparse_coo_tensor',
    'sparse_csc_tensor',
    'sparse_csr_tensor',
    'tensor',
    'tril_indices',
    'triu_indices',
    'zeros',
    'zeros_like',
]


TO_CUDA = (torch.device('cuda'), None, False, None)

_tensor__deepcopy__ = torch.Tensor.__deepcopy__
_tensor_to         = torch.Tensor.to
_tensor_cuda       = torch.Tensor.cuda
_tensor_cpu        = torch.Tensor.cpu
_torch_generics    = {name: getattr(torch, name) for name in GENERIC_METHOD_NAMES}
_cuda_init         = torch._C._cuda_init
_cuda_available      = torch.cuda.is_available
_cuda_device_count   = torch.cuda.device_count
_cuda_current_device = torch.cuda.current_device
_cuda_mem_get_info   = torch.cuda.mem_get_info
_cuda_get_device_capability   = torch.cuda.get_device_capability
_cuda_get_device_properties   = torch.cuda.get_device_properties
_cuda_get_device_name         = torch.cuda.get_device_name

TensorToArgs = Tuple[Optional[torch.device], Optional[torch.dtype], bool, Optional[torch.memory_format]]

to_ops: dict[torch.Tensor, TensorToArgs] = WeakTensorKeyDictionary() # type: ignore

def _tensor_new_register(*args, **kwargs):
    new_tensor: torch.Tensor = torch._C._TensorBase.__new__(*args, **kwargs)
    if (base_tensor := new_tensor._base) is not None:
        if base_tensor in to_ops:
            to_ops[new_tensor] = to_ops[base_tensor]
    return new_tensor

def _tensor_deepcopy_register(self: torch.Tensor, memo):
    new_tensor = _tensor__deepcopy__(self, memo)
    if isinstance(new_tensor, torch.Tensor):
        if self in to_ops:
            to_ops[new_tensor] = to_ops[self]
    return new_tensor

@property
def _tensor_device_property(self: torch.Tensor):
    if self in to_ops:
        return torch.device(type='cuda', index=0)
    del torch.Tensor.device
    try:
        return self.device
    finally:
        torch.Tensor.device = _tensor_device_property # type: ignore

@property
def _tensor_dtype_property(self: torch.Tensor):
    if self in to_ops:
        if (to_dtype := to_ops[self][1]) is not None:
            return to_dtype
    del torch.Tensor.dtype
    try:
        return self.dtype
    finally:
        torch.Tensor.dtype = _tensor_dtype_property # type: ignore

def _to_op_register(self: torch.Tensor, *args, **kwargs):
    parsed = torch._C._nn._parse_to(*args, **kwargs) # pyright: ignore [reportAttributeAccessIssue]
    device, dtype, *_ = parsed
    try:
        to_args = to_ops.pop(self)
    except KeyError:
        to_args = None
    if device is None: # pyright: ignore [reportUnnecessaryComparison]
        if to_args is not None:
            to_ops[self] = (to_args[0], dtype, *to_args[2:])
            return self
        return _tensor_to(self, *args, **kwargs)
    if device.type != 'cuda':
        if to_args is not None:
            if (to_dtype := to_args[1]) is not None:
                kwargs = {'dtype': to_dtype, **kwargs}
        return _tensor_to(self, *args, **kwargs)
    to_ops[self] = parsed
    return self

def _cuda_op_arg_check(device: torch.device | int | str | None) -> bool:
    if device is None:
        return True
    if isinstance(device, int):
        return True
    if isinstance(device, str):
        device = torch.device(device)
    return device.type == 'cuda'

def _cuda_op_register(self: torch.Tensor, device: torch.device | int | str | None = None, **kwargs):
    if not _cuda_op_arg_check(device):
        # Let PyTorch handle the fail
        return _tensor_cuda(self, device, **kwargs)
    to_ops[self] = TO_CUDA
    return self

def _cpu_op_remove(self: torch.Tensor, **kwargs):
    try:
        to_args = to_ops.pop(self)
    except KeyError:
        to_args = None
    if to_args is not None:
        if (to_dtype := to_args[1]) is not None:
            return _tensor_to(self, 'cpu', **{'dtype': to_dtype, **kwargs})
    return _tensor_cpu(self, **kwargs)

def _cuda_init_raise():
    raise RuntimeError(
        "CUDA must not be initialized in the main process "
        "on Spaces with Stateless GPU environment.\n"
        "You can look at this Stacktrace to find out "
        "which part of your code triggered a CUDA init"
    )

def _generic_method_register(name: str, *args: Any, **kwargs: Any):
    try:
        device = torch.device(kwargs.get('device', "cpu"))
    except Exception:
        return _torch_generics[name](*args, **kwargs)
    if device.type != 'cuda':
        return _torch_generics[name](*args, **kwargs)
    tensor = _torch_generics[name](*args, **{**kwargs, 'device': "cpu"})
    to_ops[tensor] = TO_CUDA
    return tensor

def patch():
    torch.Tensor.__deepcopy__ = _tensor_deepcopy_register
    torch.Tensor.__new__      = _tensor_new_register # pyright: ignore [reportAttributeAccessIssue]
    torch.Tensor.to         = _to_op_register   # type: ignore
    torch.Tensor.cuda       = _cuda_op_register # type: ignore
    torch.Tensor.cpu        = _cpu_op_remove # type: ignore
    if Config.zero_patch_torch_device:
        torch.Tensor.device = _tensor_device_property # type: ignore
        torch.Tensor.dtype  = _tensor_dtype_property # pyright: ignore [reportAttributeAccessIssue]
    for name in GENERIC_METHOD_NAMES:
        setattr(torch, name, partial(_generic_method_register, name))
    torch._C._cuda_init     = _cuda_init_raise
    torch.cuda.is_available   = lambda: True
    torch.cuda.device_count   = lambda: 1
    torch.cuda.current_device = lambda: 0
    torch.cuda.mem_get_info   = lambda *args, **kwargs: CUDA_MEM_GET_INFO
    torch.cuda.get_device_capability = lambda *args, **kwargs: CUDA_DEVICE_CAPABILITY
    torch.cuda.get_device_properties = lambda *args, **kwargs: CUDA_DEVICE_PROPERTIES
    torch.cuda.get_device_name       = lambda *args, **kwargs: CUDA_DEVICE_NAME
    bitsandbytes.patch()

def unpatch():
    torch.Tensor.__deepcopy__ = _tensor__deepcopy__
    with suppress(AttributeError):
        del torch.Tensor.__new__
    torch.Tensor.to         = _tensor_to
    torch.Tensor.cuda       = _tensor_cuda
    torch.Tensor.cpu        = _tensor_cpu
    with suppress(AttributeError):
        del torch.Tensor.device
    with suppress(AttributeError):
        del torch.Tensor.dtype
    for name in GENERIC_METHOD_NAMES:
        setattr(torch, name, _torch_generics[name])
    torch._C._cuda_init     = _cuda_init
    torch.cuda.is_available   = _cuda_available
    torch.cuda.device_count   = _cuda_device_count
    torch.cuda.current_device = _cuda_current_device
    torch.cuda.mem_get_info   = _cuda_mem_get_info
    torch.cuda.get_device_capability = _cuda_get_device_capability
    torch.cuda.get_device_properties = _cuda_get_device_properties
    torch.cuda.get_device_name       = _cuda_get_device_name
    bitsandbytes.unpatch()

def pack():
    pass

def init(nvidia_uuid: str):
    os.environ['CUDA_VISIBLE_DEVICES'] = nvidia_uuid
    torch.Tensor([0]).cuda() # CUDA init

def size():
    return 0

def move(callback: Callable[[int]] | None = None):
    for op in to_ops.items():
        tensor, parsed_args = op
        _, dtype, _, memory_format = parsed_args
        tensor.data = _tensor_to(tensor,
            device='cuda',
            dtype=dtype,
            memory_format=memory_format,
        ) # type: ignore
    bitsandbytes.move()
    torch.cuda.synchronize()

def is_in_bad_fork():
    with ProcessPoolExecutor(mp_context=multiprocessing.get_context('fork')) as e:
        f = e.submit(torch.cuda._is_in_bad_fork)
        return f.result()

def disable_cuda_intercept():
    torch.Tensor.to   = _tensor_to
    torch.Tensor.cuda = _tensor_cuda



"""
"""

from ..config import Config


if Config.zero_gpu:

    from . import client
    from . import decorator
    from . import gradio
    from . import torch
    from . import utils

    if torch.is_in_bad_fork():
        raise RuntimeError(
            "CUDA has been initialized before importing the `spaces` package. "
            "Try importing `spaces` before any other CUDA-related package."
        )

    def startup():
        total_size = torch.pack()
        if len(decorator.decorated_cache) == 0:
            return # pragma: no cover
        if Config.zerogpu_size == 'auto':
            gpu_size = 'medium' if total_size < Config.zerogpu_medium_size_threshold else 'large'
        else:
            gpu_size = Config.zerogpu_size
        client.startup_report(utils.self_cgroup_device_path(), gpu_size)

    torch.patch()
    gradio.one_launch(startup)
